# -*- coding: utf-8 -*-

# Copyright (C) 2012-2018 Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#

from __future__ import absolute_import
from __future__ import unicode_literals

import itertools
import logging

import dnf.exceptions

import tests.support


class CommonTest(tests.support.ResultTestCase):

    """Tests common to any 'multilib_policy' and 'best'.

    The test fixture consists of a dnf.Base instance that:

    - contains a package "lotus-3-17.x86_64" (The package can be installed.)
    - contains a package "lotus-3-17.i686" (The package can be installed.)
    - contains a package "trampoline-2.1-1.noarch" that contains
      "/all/be/there", provides "splendid > 2.0" and "php(a/b)" (The package
      can be installed.)
    - contains a package "mrkite-2-0.x86_64" (The package can be installed
      together with the package "trampoline".)
    - contains a package "mrkite-k-h-1-1.x86_64" (The package can be
      installed.)
    - contains a package "pepper-20-0.src"
    - contains a package "pepper-20-2.x86_64" (The package cannot be
      installed.)
    - contains a package "librita-1-1.x86_64" (The package is already
      installed.)
    - contains a package "hole-1-2.x86_64" (The package can be installed as an
      upgrade.)

    """

    REPOS = ['main', 'third_party', 'broken_deps']

    def test_install_arch_glob(self):
        """Test that the pkg specification can contain an architecture glob."""
        self.base.install("lotus.*6*")
        installed = self.installed_removed(self.base)[0]
        self.assertCountEqual(map(str, installed),
                              ['lotus-3-17.i686',
                               'lotus-3-17.x86_64'])

    def test_install_filename_glob(self):
        """Test that the pkg to be installed can be specified by fname glob."""
        self.base.install("*/there")
        (installed, _) = self.installed_removed(self.base)
        self.assertCountEqual(map(str, installed),
                              ('trampoline-2.1-1.noarch',))

        self.base.install("/all/*/there")
        (installed, _) = self.installed_removed(self.base)
        self.assertCountEqual(map(str, installed),
                              ('trampoline-2.1-1.noarch',))

    def test_install_name(self):
        """Test that the package to be installed can be specified by name."""
        self.base.install("mrkite")
        available = self.base.sack.query().available()
        expected = available.filter(name=["mrkite", "trampoline"]).run()
        # ensure sanity of the test (otherwise it would pass no matter what):
        self.assertEqual(len(expected), 2)
        new_set = self.base.sack.query().installed() + expected
        self.assertResult(self.base, new_set)

    def test_install_name_glob(self):
        """Test that the pkg to be installed can be specified by name glob."""
        self.base.install("mrkite*")
        installed = self.installed_removed(self.base)[0]
        self.assertCountEqual(map(str, installed),
                              ['mrkite-2-0.x86_64',
                               'mrkite-k-h-1-1.x86_64',
                               'trampoline-2.1-1.noarch'])

    def test_install_name_glob_exclude(self):
        """Test that glob excludes play well with glob installs."""
        subj_ex = dnf.subject.Subject('*-1')
        pkgs_ex = subj_ex.get_best_query(self.base.sack)
        self.base.sack.add_excludes(pkgs_ex)

        self.base.install('mrkite*')
        (installed, _) = self.installed_removed(self.base)
        self.assertCountEqual(map(str, installed),
                              ['mrkite-2-0.x86_64',
                               'trampoline-2.1-1.noarch'])

    def test_install_nevra(self):
        """Test that the package to be installed can be specified by NEVRA."""
        self.base.install("lotus-3-17.i686")
        lotus, = dnf.subject.Subject('lotus-3-17.i686').get_best_query(self.base.sack)
        new_set = self.base.sack.query().installed() + [lotus]
        self.assertResult(self.base, new_set)

    def test_install_provide_slash(self):
        self.base.install("php(a/b)")
        (installed, _) = self.installed_removed(self.base)
        self.assertCountEqual(map(str, installed),
                              ('trampoline-2.1-1.noarch',))

    def test_install_provide_version(self):
        """Test that the pkg to be installed can be spec. by provide ver."""
        self.base.install('splendid > 2.0')
        self.assertGreater(self.base._goal.req_length(), 0)
        trampoline = self.base.sack.query().available().filter(
            name="trampoline")
        new_set = self.base.sack.query().installed() + trampoline.run()
        self.assertResult(self.base, new_set)

    def test_package_install_conflict(self):
        """Test that a conflicting package cannot be installed."""
        p = self.base.sack.query().available().filter(
            nevra='pepper-20-2.x86_64')[0]
        self.assertEqual(1, self.base.package_install(p))
        with self.assertRaises(dnf.exceptions.DepsolveError):
            self.base.resolve()

    def test_package_install_installed(self):
        """Test that nothing changes if an installed package matches."""
        p = self.base.sack.query().available()._nevra("librita-1-1.x86_64")[0]
        with tests.support.mock.patch('logging.Logger.warning'):
            self.base.package_install(p)
        self.base.resolve()
        self.assertEmpty(self.base._goal.list_reinstalls())

        self.base.history.close()

        self.base.package_reinstall(p)
        self.base.resolve()
        self.assertLength(self.base._goal.list_reinstalls(), 1)

    def test_package_install_upgrade(self):
        """Test that the pkg to be installed can be an upgrade."""
        p = self.base.sack.query().available().filter(
            nevra='hole-1-2.x86_64')[0]
        self.assertEqual(1, self.base.package_install(p))
        installed, removed = self.installed_removed(self.base)
        self.assertIn(p, installed)
        self.assertGreaterEqual(
            removed,
            set(self.base.sack.query().installed().filter(name='hole')))

    def test_pkg_install_installable(self):
        """Test that the package to be installed can be a package instance."""
        p = self.base.sack.query().available().filter(
            nevra='lotus-3-17.x86_64')[0]
        self.assertEqual(1, self.base.package_install(p))
        self.base.resolve()
        self.assertEqual(1, len(self.base._goal.list_installs()))

    def test_pkg_install_installonly(self):
        """Test that installonly packages are installed, not upgraded."""
        self.base.conf.installonlypkgs += ['hole']
        p = self.base.sack.query().available().filter(
            nevra='hole-1-2.x86_64')[0]
        self.assertEqual(1, self.base.package_install(p))
        installed, removed = self.installed_removed(self.base)
        self.assertIn(p, installed)
        self.assertFalse(
            set(removed) &
            set(self.base.sack.query().installed().filter(name='hole')))


class MultilibAllTest(tests.support.ResultTestCase):

    """Tests for multilib_policy='all'.

    The test fixture consists of a dnf.Base instance that:

    - has conf.multilib_policy == "all"
    - has conf.best == False
    - contains a package "pepper-20-2" (The package cannot be installed.)
    - contains a package "lotus-3-16.x86_64" that contains "/usr/lib*/liblot*"
      in a "main" repository (The package can be installed.)
    - contains a package "lotus-3-16.i686" that contains "/usr/lib*/liblot*" in
      the "main" repository (The package can be installed.)
    - contains a package "librita-1-1.x86_64" (The package is already
      installed.)
    - contains a package "hole-1-2.x86_64" (The package can be installed as an
      upgrade.)
    - contains a package "pepper-20-0.src"
    - contains a package "pepper-20-0.x86_64" (The package is already
      installed.)
    - contains nothing that matches "not-available"
    - contains a package that provides "henry(the_horse)" (The package can be
      installed.)
    - contains a package "lotus-3-17.x86_64" not in a "main" repository (The
      package can be installed.)
    - contains a package "lotus-3-17.i686" not in the "main" repository (The
      package can be installed.)

    """

    REPOS = ['main', 'third_party', 'broken_deps']

    def setUp(self):
        super(MultilibAllTest, self).setUp()
        self.base.conf.multilib_policy = "all"
        assert self.base.conf.best is False

    def test_install_conflict(self):
        """Test that the exception is raised if the package conflicts."""
        self.base.install('pepper-20-2')
        with self.assertRaises(dnf.exceptions.DepsolveError):
            self.base.resolve()

    def test_install_filename(self):
        """Test that the pkg to be installed can be specified by filename."""
        self.base.install("/usr/lib*/liblot*")
        inst, _ = self.installed_removed(self.base)
        self.assertCountEqual(
            map(str, inst), ['lotus-3-16.x86_64', 'lotus-3-16.i686'])

    def test_install_installed(self):
        """Test that nothing changes if an installed package matches."""
        stdout = dnf.pycomp.StringIO()
        with tests.support.wiretap_logs('dnf', logging.INFO, stdout):
            self.base.install('librita')
        self.assertEqual(self.base._goal.req_length(), 0)
        self.assertIn(
            'Package librita-1-1.x86_64 is already installed.',
            stdout.getvalue())

    def test_install_installonly(self):
        """Test that installonly packages are installed, not upgraded."""
        self.base.conf.installonlypkgs += ['hole']
        self.base.install('hole-1-2')
        installed, removed = self.installed_removed(self.base)
        self.assertGreaterEqual(
            installed,
            set(self.base.sack.query().available()._nevra('hole-1-2.x86_64')))
        self.assertEmpty(removed)

    def test_install_multilib(self):
        """Test that pkgs for all architectures are installed if available."""
        cnt = self.base.install("lotus")
        self.assertEqual(cnt, 2)
        installed = self.installed_removed(self.base)[0]
        self.assertLessEqual(
            set(installed),
            set(self.base.sack.query().available().filter(name='lotus')))

    def test_install_name_choice(self):
        """Test that the matching pkg that can be installed is installed."""
        # Don't install the SRPM.
        self.base.sack.add_excludes(
            dnf.subject.Subject('pepper.src').get_best_query(self.base.sack))

        stdout = dnf.pycomp.StringIO()
        with tests.support.wiretap_logs('dnf', logging.INFO, stdout):
            self.base.install('pepper')
        self.assertEqual(self.base._goal.req_length(), 0)
        self.assertIn(
            'Package pepper-20-0.x86_64 is already installed.',
            stdout.getvalue())

    def test_install_nonexistent(self):
        """Test that the exception is raised if no package matches."""
        with self.assertRaises(dnf.exceptions.MarkingError) as context:
            self.base.install('not-available')
        self.assertEqual(context.exception.pkg_spec, 'not-available')
        installed_pkgs = self.base.sack.query().installed().run()
        self.assertResult(self.base, installed_pkgs)

    def test_install_provide(self):
        """Test that the pkg to be installed can be specified by provide."""
        self.base.install("henry(the_horse)")
        self.assertGreater(self.base._goal.req_length(), 0)

    def test_install_reponame(self):
        """Test whether packages are filtered by the reponame."""
        result = itertools.chain(
            self.base.sack.query().installed(),
            dnf.subject.Subject('lotus-3-16').get_best_query(self.base.sack))

        self.base.install('lotus', reponame='main')
        self.assertResult(self.base, result)

        assert dnf.subject.Subject('lotus-3-17').get_best_query(self.base.sack), \
            ('the base must contain packages a package in another repo '
             'which matches the pattern but is preferred, otherwise the '
             'test makes no sense')

    def test_install_upgrade(self):
        """Test that the pkg to be installed can be an upgrade."""
        self.base.install('hole-1-2')
        installed, removed = self.installed_removed(self.base)
        self.assertGreaterEqual(
            installed,
            set(self.base.sack.query().available()._nevra('hole-1-2.x86_64')))
        self.assertGreaterEqual(
            removed,
            set(self.base.sack.query().installed().filter(name='hole')))


class MultilibBestTest(tests.support.ResultTestCase):

    """Tests for multilib_policy='best'.

    The test fixture consists of a dnf.Base instance that:

    - has conf.multilib_policy == "best"
    - has conf.best == False
    - contains a package "pepper-20-2" (The package cannot be installed.)
    - contains a package "lotus-3-16.x86_64" that contains "/usr/lib*/liblot*"
      in a "main" repository (The package can be installed.)
    - contains a package "lotus-3-16.i686" that contains "/usr/lib*/liblot*"
      in the "main" repository (The package can be installed.)
    - contains nothing that matches "/not/exist/"
    - contains a package "librita-1-1.x86_64" (The package is already
      installed.)
    - contains a package "hole-1-2.x86_64" (The package can be installed as an
      upgrade.)
    - contains a package "lotus-3-17.x86_64" not in the "main" repository (The
      package can be installed.)
    - contains a package "pepper-20-0.src"
    - contains a package "pepper-20-0.x86_64" (The package is already
      installed.)
    - contains nothing that matches "not-available"
    - contains a package "trampoline" that provides "henry(the_horse)" (The
      package can be installed.)
    - contains a package "lotus-3-17.i686" not in the "main" repository (The
      package can be installed.)
    - contains a package "hole-1-1.x86_64" (The package is already installed
      and is not available.)

    """

    REPOS = ['main', 'third_party', 'broken_deps']

    def setUp(self):
        super(MultilibBestTest, self).setUp()
        self.installed = self.base.sack.query().installed().run()
        self.assertEqual(self.base.conf.multilib_policy, "best")
        assert self.base.conf.best is False

    def test_install_conflict(self):
        """Test that the exception is raised if the package conflicts."""
        self.base.install('pepper-20-2')
        with self.assertRaises(dnf.exceptions.DepsolveError):
            self.base.resolve()

    def test_install_filename(self):
        """Test that the pkg to be installed can be specified by filename."""
        self.base.install("/usr/lib*/liblot*")
        inst, _ = self.installed_removed(self.base)
        self.assertCountEqual(map(str, inst), ['lotus-3-16.x86_64'])

        self.assertRaises(dnf.exceptions.MarkingError,
                          self.base.install, "/not/exist/")

    def test_install_installed(self):
        """Test that nothing changes if an installed package matches."""
        stdout = dnf.pycomp.StringIO()
        with tests.support.wiretap_logs('dnf', logging.INFO, stdout):
            self.base.install('librita')
        installed, removed = self.installed_removed(self.base)
        self.assertEmpty(installed)
        self.assertEmpty(removed)
        self.assertIn(
            'Package librita-1-1.x86_64 is already installed.',
            stdout.getvalue())

    def test_install_installonly(self):
        """Test that installonly packages are installed, not upgraded."""
        self.base.conf.installonlypkgs += ['hole']
        self.base.install('hole-1-2')
        installed, removed = self.installed_removed(self.base)
        self.assertGreaterEqual(
            installed,
            set(self.base.sack.query().available()._nevra('hole-1-2.x86_64')))
        self.assertEmpty(removed)

    def test_install_multilib(self):
        """Test that a pkg for only one architecture are installed."""
        cnt = self.base.install("lotus")
        self.assertEqual(cnt, 1)

        new_package, = dnf.subject.Subject('lotus-3-17.x86_64').get_best_query(self.base.sack)
        new_set = self.installed + [new_package]
        self.assertResult(self.base, new_set)

    def test_install_name_choice(self):
        """Test that the matching pkg that can be installed is installed."""
        # Don't install the SRPM.
        self.base.sack.add_excludes(
            dnf.subject.Subject('pepper.src').get_best_query(self.base.sack))

        stdout = dnf.pycomp.StringIO()
        with tests.support.wiretap_logs('dnf', logging.INFO, stdout):
            self.base.install('pepper')
        installed, removed = self.installed_removed(self.base)
        self.assertEmpty(installed | removed)
        self.assertIn(
            'Package pepper-20-0.x86_64 is already installed.',
            stdout.getvalue())

    def test_install_nonexistent(self):
        """Test that the exception is raised if no package matches."""
        with self.assertRaises(dnf.exceptions.MarkingError) as context:
            self.base.install('not-available')
        self.assertEqual(context.exception.pkg_spec, 'not-available')
        installed_pkgs = self.base.sack.query().installed().run()
        self.assertResult(self.base, installed_pkgs)

    def test_install_provide(self):
        """Test that the pkg to be installed can be specified by provide."""
        self.base.install("henry(the_horse)")
        self.assertGreater(self.base._goal.req_length(), 0)
        trampoline = self.base.sack.query().available().filter(
            name="trampoline")
        new_set = self.installed + trampoline.run()
        self.assertResult(self.base, new_set)

    def test_install_reponame(self):
        """Test whether packages are filtered by the reponame."""
        result = itertools.chain(
            self.base.sack.query().installed(),
            dnf.subject.Subject('lotus-3-16.x86_64')
            .get_best_query(self.base.sack))

        self.base.install('lotus', reponame='main')
        self.assertResult(self.base, result)

        assert dnf.subject.Subject('lotus-3-17.x86_64').get_best_query(self.base.sack), \
            ('the base must contain packages a package in another repo '
             'which matches the pattern but is preferred, otherwise the '
             'test makes no sense')

    def test_install_unavailable(self):
        """Test that nothing changes if an unavailable package matches."""
        stdout = dnf.pycomp.StringIO()
        with tests.support.wiretap_logs('dnf', logging.INFO, stdout):
            cnt = self.base.install('hole')
        self.assertEqual(cnt, 1)
        installed_pkgs = self.base.sack.query().installed().run()
        self.assertResult(self.base, installed_pkgs)
        self.assertIn(
            'Package hole-1-1.x86_64 is already installed.',
            stdout.getvalue())

    def test_install_upgrade(self):
        """Test that the pkg to be installed can be an upgrade."""
        self.base.install('hole-1-2')
        installed, removed = self.installed_removed(self.base)
        self.assertGreaterEqual(
            installed,
            set(self.base.sack.query().available()._nevra('hole-1-2.x86_64')))
        self.assertGreaterEqual(
            removed,
            set(self.base.sack.query().installed().filter(name='hole')))


class BestTrueTest(tests.support.ResultTestCase):

    """Tests for best=True.

    The test fixture consists of a dnf.Base instance that:

    - has conf.best == True
    - contains a package "pepper-20-2" (The package cannot be installed.)

    """

    REPOS = ['broken_deps']

    def setUp(self):
        super(BestTrueTest, self).setUp()
        self.base.conf.best = True

    def test_install_name_choice(self):
        """Test that the latest version of the matching pkg is installed."""
        with tests.support.mock.patch('logging.Logger.warning'):
            self.base.install('pepper')
        with self.assertRaises(dnf.exceptions.DepsolveError):
            self.base.resolve()



# Copyright (C) 2015  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.

"""A plugin that lists installed packages that are obsolted by any available package"""

from dnf.i18n import _
import dnf
import dnf.cli


# If you only plan to create a new dnf subcommand in a plugin
# you can use @dnf.plugin.register_command decorator instead of creating
# a Plugin class which only registers the command
# (for full-fledged Plugin class see examples/install_plugin.py)
@dnf.plugin.register_command
class Command(dnf.cli.Command):

    """A command that lists packages installed on the system that are
       obsoleted by packages in any known repository."""

    # An alias is needed to invoke the command from command line.
    aliases = ['foo']  # <-- SET YOUR ALIAS HERE.

    @staticmethod
    def set_argparser(parser):
        parser.add_argument("package", nargs='*', metavar=_('PACKAGE'))

    def configure(self):
        """Setup the demands."""
        # Repositories serve as sources of information about packages.
        self.cli.demands.available_repos = True
        # A sack is needed for querying.
        self.cli.demands.sack_activation = True

    def run(self):
        """Run the command."""

        obs_tuples = []
        # A query matches all available packages
        q = self.base.sack.query()

        if not self.opts.package:
            # Narrow down query to match only installed packages
            inst = q.installed()
            # A dictionary containing list of obsoleted packages
            for new in q.filter(obsoletes=inst):
                obs_reldeps = new.obsoletes
                obsoleted = inst.filter(provides=obs_reldeps).run()
                obs_tuples.extend([(new, old) for old in obsoleted])
        else:
            for pkg_spec in self.opts.package:
                # A subject serves for parsing package format from user input
                subj = dnf.subject.Subject(pkg_spec)
                # A query restricted to installed packages matching given subject
                inst = subj.get_best_query(self.base.sack).installed()
                for new in q.filter(obsoletes=inst):
                    obs_reldeps = new.obsoletes
                    obsoleted = inst.filter(provides=obs_reldeps).run()
                    obs_tuples.extend([(new, old) for old in obsoleted])

        if not obs_tuples:
            raise dnf.exceptions.Error('No matching Packages to list')

        for (new, old) in obs_tuples:
            print('%s.%s obsoletes %s.%s' %
                  (new.name, new.arch, old.name, old.arch))





# -*- coding: utf-8 -*-

# Copyright (C) 2019 Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#

from __future__ import absolute_import
from __future__ import unicode_literals

import os

import dnf.conf

import tests.support


FN = tests.support.resource_path('etc/repos.conf')


class SubstitutionsFromEnvironmentTest(tests.support.TestCase):
    def test_numeric(self):
        env = os.environ
        os.environ['DNF0'] = 'the_zero'
        conf = dnf.conf.Conf()
        os.environ = env
        self.assertIn('DNF0', conf.substitutions)
        self.assertEqual('the_zero', conf.substitutions['DNF0'])

    def test_named(self):
        env = os.environ
        os.environ['DNF_VAR_GENRE'] = 'opera'
        os.environ['DNF_VAR_EMPTY'] = ''
        os.environ['DNF_VAR_MAL$FORMED'] = 'not this'
        os.environ['DNF_VARMALFORMED'] = 'not this'
        os.environ['DNF_VAR_MALFORMED '] = 'not this'
        conf = dnf.conf.Conf()
        os.environ = env
        self.assertCountEqual(
            conf.substitutions.keys(),
            ['basearch', 'arch', 'GENRE', 'EMPTY'])
        self.assertEqual('opera', conf.substitutions['GENRE'])



#!/usr/bin/python2
#
# Copyright (C) 2015 Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#

r"""A script intended to update release notes.

Works on Python >= 2.7 and < 3.0.

This module, when run as a script, provides a command line interface.
The interface usage is::

    usage: prog [-h]

    Extend the list of issues of the next version in release notes. It
    updates the release notes found in the DNF repository which contains
    this module. The module location is obtained from the __file__
    variable. The version number is parsed from the spec file. If the
    release notes already contains given issue, it is not added again.
    Otherwise, it is added to the beginning of the list. If the release
    notes does not contain the next version, a new section is appended
    right before the first section. The diff between the new release
    notes and the Git index is printed to the standard error stream. The
    message starts with "INFO Update finished. See the diff to the Git
    index: ".

    optional arguments:
      -h, --help  show this help message and exit

    If the version is already released or if another error occurs the
    exit status is non-zero.

:var LOGGER: the logger used by this script
:type LOGGER: logging.Logger
:var SPEC_FN: a name of the DNF spec file relative to the working
   directory of the Git repository which contains this script
:type SPEC_FN: str
:var TAG_PAT: a string pattern of DNF tags where the "{version}" field
   should be replaced with the given version of DNF
:type TAG_PAT: str
:var ISSUE_RE: a regular expression matching numbers of resolved issues
   in the first line of a commit message
:type ISSUE_RE: re.RegexObject
:var RELATED_RE: a regular expression matching numbers of related issues
   in a commit message
:type RELATED_RE: re.RegexObject
:var SIMILAR_RE: a regular expression matching numbers which look like a
   number of an issue referred in a commit message
:type SIMILAR_RE: re.RegexObject
:var NOTES_FN: a name of the DNF release notes file relative to the
   working directory of the Git repository which contains this script
:type NOTES_FN: str
:var TITLE_PAT: a string pattern of a DNF release notes section title
   (including a trailing "\n") where the "{version}" field should be
   replaced with the given release version
:type TITLE_PAT: str
:var ISSUE_PAT: a string pattern of a reference to a resolved issue in
   DNF release notes (including a trailing "\n") where the "{number}"
   field should be replaced with the number of the given issue
:type ISSUE_PAT: str
:var OVERLINE_RE: a regular expression matching a section title overline
   in a line of DNF release notes
:type OVERLINE_RE: re.RegexObject
:var TITLE_RE_PAT: a string pattern of a regular expression matching a
   section title in a line of DNF release notes where the "{maxlen}"
   field should be replaced with the maximum expected length of the
   release version
:type TITLE_RE_PAT: str

"""


from __future__ import absolute_import

import argparse
import contextlib
import io
import itertools
import logging
import os
import re
import shutil
import sys
import tempfile
import unittest

import git
import rpm
import tests.mock


LOGGER = logging.getLogger(sys.argv[0])

SPEC_FN = 'dnf.spec'

TAG_PAT = 'dnf-{version}-1'

ISSUE_RE = re.compile(ur'(?<!Related: )RhBug:(\d+)')

RELATED_RE = re.compile(ur'Related: RhBug:(\d+)')

SIMILAR_RE = re.compile(ur'\d{6,}')

NOTES_FN = os.path.join('doc', 'release_notes.rst')

TITLE_PAT = '{version} Release Notes\n'

ISSUE_PAT = '* :rhbug:`{number}`\n'

OVERLINE_RE = re.compile('^(=+)\n$')

TITLE_RE_PAT = '^(.{{,{maxlen}}}) Release Notes\n$'


@contextlib.contextmanager
def _safe_overwrite(name):
    """Open a file in order to overwrite it safely.

    Until the context manager successfully exits, the file remains
    unchanged.

    :param name: name of the file
    :type name: str
    :returns: a context manager that returns a readable text file object
        intended to read the content of the file and a writable text
        file object intended to write the new content of the file
    :rtype: contextmanager[tuple[file, file]]
    :raises exceptions.IOError: if the file cannot be opened

    """
    with open(name, 'rt+') as file_, tempfile.TemporaryFile('wt+') as tmpfile:
        yield file_, tmpfile
        tmpfile.seek(0)
        file_.seek(0)
        shutil.copyfileobj(tmpfile, file_)


def detect_repository():
    """Detect the DNF Git repository which contains this module.

    The module location is obtained from :const:`__file__`.

    :returns: the repository
    :rtype: git.Repo

    """
    dirname = os.path.dirname(__file__)
    # FIXME remove this once we can get rid of supporting GitPython < 0.3.5
    try:
        return git.Repo(dirname, search_parent_directories=True)
    except TypeError:
        return git.Repo(dirname)


def detect_version(repository):
    """Detect the DNF version from its spec file.

    :param repository: a non-bare DNF repository which contains the spec
       file
    :type repository: git.Repo
    :returns: the detected version
    :rtype: str

    """
    filename = os.path.join(repository.working_dir, SPEC_FN)
    return rpm.spec(filename).sourceHeader[rpm.RPMTAG_VERSION]


def find_tag(repository, version):
    """Find the Git tag corresponding to a DNF version.

    :param repository: a DNF repository
    :type repository: git.Repo
    :returns: the name of the tag
    :rtype: str | None

    """
    tagname = TAG_PAT.format(version=version)
    return tagname if tagname in repository.tags else None


def iter_unreleased_commits(repository):
    """Iterate over the commits that were not tagged as a release.

    :param repository: a repository
    :type repository: git.Repo
    :returns: a generator yielding the commits
    :rtype: generator[git.Commit]

    """
    tagshas = {tag.commit.hexsha for tag in repository.tags}
    for commit in repository.iter_commits():
        if commit.hexsha in tagshas:
            # FIXME encode() once we get rid of supporting GitPython < 0.3.4
            LOGGER.debug(
                'Unreleased commits iteration stopped on the first tagged '
                'commit: %s', commit.hexsha.encode())
            break
        yield commit


def parse_issues(commits):
    """Parse the numbers of the resolved DNF issues from commit messages.

    :param commits: the DNF commits
    :type commits: collections.Iterable[git.Commit]
    :returns: a generator yielding the issue numbers
    :rtype: generator[str]

    """
    for commit in commits:
        firstline = commit.message.splitlines()[0]
        valid = {match.group(1) for match in ISSUE_RE.finditer(firstline)}
        for issue in valid:
            issue = issue.encode()
            # FIXME decode() once we get rid of supporting GitPython < 0.3.4
            LOGGER.debug(
                'Recognized %s in commit %s.', issue,
                commit.hexsha.decode().encode())
            yield issue.encode()
        valid |= {mat.group(1) for mat in RELATED_RE.finditer(commit.message)}
        for match in SIMILAR_RE.finditer(commit.message):
            if match.group(0) not in valid:
                # FIXME decode once we get rid of supporting GitPython < 0.3.4
                LOGGER.warning(
                    'Skipped %s in commit %s which looks like an issue '
                    'number.', match.group(0).encode(),
                    commit.hexsha.decode().encode())


def extend_releases(releases, version, issues):
    r"""Extend the list of issues of a release version in release notes.

    The issues are appended to the beginning of the list. If an issue is
    already in the list of the release version, it is not added again.
    If the release version is not found, the release version
    (``version, '\n', issues, '\n'``) is yielded right before the first
    release version in the release notes.

    :param releases: the version, the description, the list of issue
       numbers and the epilog for each release version in release notes
    :type releases: collections.Iterable[tuple[str | None, str, list[str], str]]
    :param version: the release version to be extended
    :type version: str
    :param issues: the issues to be added
    :type issues: collections.Iterable[str]
    :returns: a generator yielding the modified release notes
    :rtype:  generator[tuple[str | None, str, list[str], str]]

    """
    releases, issues = list(releases), list(issues)
    prepend = not any(release[0] == version for release in releases)
    for version_, description, issues_, epilog in releases:
        if prepend and version_ is not None:
            yield version, '\n', issues, '\n'
            prepend = False
        elif version_ == version:
            issues_ = issues_[:]
            for issue in reversed(issues):
                if issue not in issues_:
                    issues_.insert(0, issue)
        yield version_, description, issues_, epilog


def format_release(version, description, issues, epilog):
    """Compose a string in form of DNF release notes from a release.

    The order of issues is preserved.

    :param version: the version of the release
    :type version: str | None
    :param description: the description of the release
    :type description: str
    :param issues: the list of issue numbers of the release
    :type issues: list[str]
    :param epilog: the epilog of the release
    :type epilog: str
    :returns: the formatted string
    :rtype: str

    """
    titlestr = ''
    if version:
        title = TITLE_PAT.format(version=version)
        length = len(title) - 1
        titlestr = '{}\n{}{}\n'.format('=' * length, title, '=' * length)
    issuestr = ''.join(ISSUE_PAT.format(number=issue) for issue in issues)
    return ''.join([titlestr, description, issuestr, epilog])


def update_notes():
    r"""Extend the list of issues of the next version in release notes.

    It updates the release notes found in the DNF repository which
    contains this module. The module location is obtained from
    :const:`__file__`. The version number is parsed from the spec file.
    If the release notes already contains given issue, it is not added
    again. Otherwise, it is added to the beginning of the list. If the
    release notes does not contain the next version, a new section is
    appended right before the first section. The diff between the new
    release notes and the Git index is logged to :const:`.LOGGER` with
    level :const:`INFO`. The message starts with "Update finished. See
    the diff to the Git index:\n".

    :raises exceptions.ValueError: if the version is already released

    """
    repository = detect_repository()
    LOGGER.info('Detected the repository at: %s', repository.working_dir)
    notesfn = os.path.join(repository.working_dir, NOTES_FN)
    version = detect_version(repository)
    LOGGER.info('Detected DNF version (from the spec file): %s', version)
    issues = parse_issues(iter_unreleased_commits(repository))
    parser = ReleaseNotesParser()
    tagname = find_tag(repository, version)
    if tagname:
        LOGGER.warning('Found a tag matching the current version: %s', tagname)
        LOGGER.error('Extending an already released version is not allowed!')
        raise ValueError('version already released')
    with _safe_overwrite(notesfn) as (source, destination):
        releases = extend_releases(parser.parse_lines(source), version, issues)
        for version_, desc, issues_, epilog in releases:
            destination.write(format_release(version_, desc, issues_, epilog))
    diff = repository.index.diff(None, NOTES_FN, create_patch=True)[0].diff
    LOGGER.info(
        'Update finished. See the diff to the Git index:\n%s',
        re.sub(r'^', '    ', diff, flags=re.M))


def main():
    """Start the main loop of the command line interface.

    The root logger is configured to write DEBUG+ messages into the
    directory where is this module located if not configured otherwise.
    A handler that writes INFO+ messages to :data:`sys.stderr` is added
    to :const:`.LOGGER`.

    The interface usage is::

        usage: prog [-h]

        Extend the list of issues of the next version in release notes.
        It updates the release notes found in the DNF repository which
        contains this module. The module location is obtained from the
        __file__ variable. The version number is parsed from the spec
        file. If the release notes already contains given issue, it is
        not added again. Otherwise, it is added to the beginning of the
        list. If the release notes does not contain the next version, a
        new section is appended right before the first section. The diff
        between the new release notes and the Git index is printed to
        the standard error stream. The message starts with "INFO Update
        finished. See the diff to the Git index: ".

        optional arguments:
          -h, --help  show this help message and exit

        If the version is already released or if another error occurs
        the exit status is non-zero.

    :raises exceptions.SystemExit: with a non-zero exit status if an
       error occurs

    """
    logging.basicConfig(
        filename='{}.log'.format(__file__),
        filemode='wt',
        format='%(asctime)s.%(msecs)03d:%(levelname)s:%(name)s:%(message)s',
        datefmt='%Y%m%dT%H%M%S',
        level=logging.DEBUG)
    handler = logging.StreamHandler()
    handler.setLevel(logging.INFO)
    handler.setFormatter(logging.Formatter('%(levelname)s %(message)s'))
    LOGGER.addHandler(handler)
    argparser = argparse.ArgumentParser(
        description='Extend the list of issues of the next version in release '
                    'notes. It updates the release notes found in the DNF '
                    'repository which contains this module. The module '
                    'location is obtained from the __file__ variable. The '
                    'version number is parsed from the spec file. If the '
                    'release notes already contains given issue, it is not '
                    'added again. Otherwise, it is added to the beginning of '
                    'the list. If the release notes does not contain the next '
                    'version, a new section is appended right before the first'
                    ' section. The diff between the new release notes and the '
                    'Git index is printed to the standard error stream. The '
                    'message starts with "INFO Update finished. See the diff '
                    'to the Git index:\n".',
        epilog='If the version is already released or if another error occurs '
               'the exit status is non-zero.')
    argparser.parse_args()
    try:
        update_notes()
    except ValueError:
        sys.exit(1)


class ReleaseNotesParser(object):

    """A class able to parse DNF release notes.

    :ivar _needline: an expected line which represents the end of a
       section title
    :type _needline: str | None
    :ivar _version: a version parsed from a potential next section title
    :type _version: str | None
    :ivar version: a version parsed from the last section title
    :type version: str | None
    :ivar description: lines parsed from the last section description
    :type description: list[str]
    :ivar issues: numbers of resolved issues parsed from the last
       section
    :type issues: list[str]
    :ivar epilog: lines parsed from the last section epilog
    :type epilog: list[str]

    """

    def __init__(self):
        """Initialize the parser."""
        super(ReleaseNotesParser, self).__init__()
        self._needline = None
        self._version = None
        self.version = None
        self.description = []
        self.issues = []
        self.epilog = []

    @property
    def _last_version(self):
        """The last parsed release version.

        :returns: the version, the description, the list of issue
           numbers and the epilog of the release
        :rtype: tuple[str | None, str, list[str], str]

        """
        return (
            self.version,
            ''.join(self.description),
            self.issues,
            ''.join(self.epilog))

    def _cancel_title(self):
        """Cancel expectations of a next section title.

        It modifies :attr:`_needline` and :attr:`_version`.

        :returns: the old values of :attr:`_needline` and
           :attr:`_version`
        :rtype: tuple[str | None, str | None]

        """
        needline, version = self._needline, self._version
        self._needline = self._version = None
        return needline, version

    def _parse_overline(self, line):
        """Parse the overline of the next section title from a line.

        It changes the state of the parser.

        :param line: the line to be parsed
        :type line: str
        :returns: returns :data:`True` if the line contains an
            overline, otherwise it returns :data:`False`
        :rtype: bool

        """
        match = OVERLINE_RE.match(line)
        if not match:
            return False
        self._cancel_title()
        self._needline = match.group(1)
        return True

    def _parse_title(self, line):
        """Parse the title from a line.

        It changes the state of the parser even if the line does not
        contain a title.

        :param line: the line to be parsed
        :type line: str
        :returns: returns :data:`True` if the line contains a title,
           otherwise it returns :data:`False`
        :rtype: bool

        """
        maxlen = len(self._needline) - len(' Release Notes')
        match = re.match(TITLE_RE_PAT.format(maxlen=maxlen), line)
        if not match:
            self._cancel_title()
            return False
        self._version = match.group(1)
        return True

    def _parse_underline(self, line):
        """Parse the underline of the next section title from a line.

        It changes the state of the parser.

        :param line: the line to be parsed
        :type line: str
        :returns: the version, the description, the list of issue
           numbers and the epilog of the previous release
        :rtype: tuple[str | None, str, list[str], str] | None

        """
        needline, version = self._cancel_title()
        if line != '{}\n'.format(needline):
            return None
        previous_version = self._last_version
        self.version = version
        self.description, self.issues, self.epilog = [], [], []
        return previous_version

    def _parse_issue(self, line):
        """Parse an issue number from a line.

        It changes the state of the parser.

        :param line: the line to be parsed
        :type line: str
        :returns: returns :data:`True` if the line refers to an issue,
           otherwise it returns :data:`False`
        :rtype: bool

        """
        match = re.match(r'^\* :rhbug:`(\d+)`\n$', line)
        if not match:
            return False
        self.issues.append(match.group(1))
        return True

    def _parse_line(self, line):
        """Parse a line of DNF release notes.

        It changes the state of the parser.

        :param line: the line to be parsed
        :type line: str
        :returns: the version, the description, the list of issue
           numbers and the epilog of the previous release
        :rtype: tuple[str | None, str, list[str], str] | None

        """
        needtitle = self._needline and self._version
        if not needtitle and self._parse_overline(line):
            return None
        if self._needline and not self._version and self._parse_title(line):
            return None
        if self._needline and self._version:
            previous_version = self._parse_underline(line)
            if previous_version:
                return previous_version
        if not self.epilog and self._parse_issue(line):
            return None
        if not self.issues:
            self.description.append(line)
            return None
        self.epilog.append(line)

    def parse_lines(self, lines):
        """Parse the lines of DNF release notes.

        :param lines: the line to be parsed
        :type lines: collections.Iterable[str]
        :returns: a generator yielding the version, the description, the
           list of issue numbers and the epilog of each release version
        :rtype: generator[tuple[str | None, str, list[str], str]]

        """
        self._needline = None
        self._version = None
        self.version = None
        self.description = []
        self.issues = []
        self.epilog = []
        for line in lines:
            previous_version = self._parse_line(line)
            if previous_version:
                yield previous_version
        yield self._last_version


class TestCase(unittest.TestCase):

    """A test fixture common to all tests.

    Among other things, the fixture contains a non-bare DNF repository
    with a spec file specifying an unreleased version, a tag dnf-1.0.1-1
    matching the version 1.0.1, one extra commit which resolves an issue
    and release notes matching the revision 9110490 of DNF.

    :cvar VERSION: the unreleased version specified in the spec file
    :type VERSION: str
    :cvar ISSUE: the number of the issue resolved by the extra commit
    :type ISSUE: str
    :ivar repository: the testing repository
    :type repository: git.Repo
    :ivar commit: the extra commit
    :type commit: unicode

    """

    VERSION = '999.9.9'

    ISSUE = '123456'

    def setUp(self):
        """Prepare the test fixture.

        :const:`__file__` is needed to prepare the fixture.

        """
        dirname = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, dirname)
        self.repository = detect_repository().clone(dirname)
        self.repository.head.reset(
            '9110490d690bbad10977b86e7ffbe1feeae26e03', working_tree=True)
        with open(os.path.join(dirname, SPEC_FN), 'wt') as specfile:
            specfile.write(
                'Name: dnf\n' +
                'Version: {}\n'.format(self.VERSION) +
                'Release: 1\n'
                'Summary: Package manager forked from Yum\n'
                'License: GPLv2+ and GPLv2 and GPL\n'
                '%description\n'
                'Package manager forked from Yum, using libsolv.\n'
                '%files\n')
        self.repository.index.add([SPEC_FN])
        self.commit = self.repository.index.commit(
            u'Version (RhBug:{})'.format(self.ISSUE.decode()))

    @staticmethod
    @contextlib.contextmanager
    def _read_copy(name):
        """Create and open a readable copy of a file.

        :param name: name of the file
        :type name: str
        :returns: the readable copy
        :rtype: file
        :raises exceptions.IOError: if the file cannot be opened

        """
        with tempfile.TemporaryFile('wt+') as copy:
            with open(name, 'rt') as original:
                shutil.copyfileobj(original, copy)
            copy.seek(0)
            yield copy

    @contextlib.contextmanager
    def _patch_file(self):
        """Temporarily set :const:`__file__` to point to the testing repo.

        :returns: a context manager
        :rtype: contextmanager

        """
        filename = os.path.join(
            self.repository.working_dir, 'scripts', 'update_releasenotes.py')
        original = __file__
        with tests.mock.patch('update_releasenotes.__file__', filename):
            assert __file__ != original, 'double check that the patch works'
            yield

    @contextlib.contextmanager
    def _assert_logs(self, level, regex):
        """Test whether a message matching an expression is logged.

        :param level: the level of the message
        :type level: int
        :param regex: the regular expression to be matched
        :type regex: re.RegexObject
        :returns: a context manager which represents the block in which
           the message should be logged
        :rtype: contextmanager
        :raises exceptions.AssertionError: if the test fails

        """
        class Handler(logging.Handler):
            def __init__(self, regex):
                super(Handler, self).__init__()
                self.regex = regex
                self.found = False

            def emit(self, record):
                if not self.found:
                    self.found = self.regex.match(record.getMessage())
        handler = Handler(regex)
        LOGGER.setLevel(level)
        LOGGER.addHandler(handler)
        yield
        self.assertTrue(handler.found)

    @contextlib.contextmanager
    def _assert_prints(self, regex, stream):
        """Test whether a message matching an expression is printed.

        :param regex: the regular expression to be matched
        :type regex: re.RegexObject
        :param stream: a name of the stream to be matched
        :type stream: str
        :returns: a context manager which represents the block in which
           the message should be printed
        :rtype: contextmanager
        :raises exceptions.AssertionError: if the test fails

        """
        with tests.mock.patch(stream, io.BytesIO()) as mock:
            yield
            self.assertRegexpMatches(mock.getvalue(), regex)

    def _assert_iter_equal(self, actual, expected):
        """Test whether two iterables are equal.

        :param actual: one of the iterables
        :type actual: collections.Iterable[object]
        :param expected: the other iterable
        :type expected: collections.Iterable[object]
        :raises exceptions.AssertionError: if the test fails

        """
        self.assertTrue(all(
            actual_ == expected_ for actual_, expected_ in
            itertools.izip_longest(actual, expected, fillvalue=object())))

    def test_detect_repository(self):
        """Test whether correct repository is detected.

        :raises exceptions.AssertionError: if the test fails

        """
        with self._patch_file():
            self.assertEqual(
                detect_repository().working_dir, self.repository.working_dir)

    def test_detect_version(self):
        """Test whether correct version is detected.

        :raises exceptions.AssertionError: if the test fails

        """
        self.assertEqual(detect_version(self.repository), self.VERSION)

    def test_find_tag_name(self):
        """Test whether correct tag is detected.

        :raises exceptions.AssertionError: if the test fails

        """
        self.assertEqual(find_tag(self.repository, '1.0.1'), 'dnf-1.0.1-1')

    def test_find_tag_none(self):
        """Test whether correct tag is detected.

        :raises exceptions.AssertionError: if the test fails

        """
        self.assertIsNone(find_tag(self.repository, '9999'))

    def test_iter_unreleased_commits(self):
        """Test whether correct commits are yielded.

        :raises exceptions.AssertionError: if the test fails

        """
        commits = iter_unreleased_commits(self.repository)
        self.assertItemsEqual(
            (commit.hexsha for commit in commits), [self.commit.hexsha])

    def test_parse_issues(self):
        """Test whether correct issues are yielded.

        :raises exceptions.AssertionError: if the test fails

        """
        self.assertItemsEqual(parse_issues([self.commit]), [self.ISSUE])

    def test_extend_releases_extend(self):
        """Test whether the release version is extended.

        :raises exceptions.AssertionError: if the test fails

        """
        releases = [
            (None, 'd1\n', [], ''),
            ('1.0.1', '\nd2\n', ['234567'], '\n'),
            ('999.9.9', '\nd3\n', ['345678'], '\n')]
        expected = [
            (None, 'd1\n', [], ''),
            ('1.0.1', '\nd2\n', ['456789', '234567'], '\n'),
            ('999.9.9', '\nd3\n', ['345678'], '\n')]
        self.assertItemsEqual(
            extend_releases(releases, '1.0.1', ['456789']), expected)

    def test_extend_releases_skip(self):
        """Test whether the already present issues are skipped.

        :raises exceptions.AssertionError: if the test fails

        """
        releases = [
            (None, 'd1\n', [], ''),
            ('1.0.1', '\nd2\n', ['234567'], '\n'),
            ('999.9.9', '\nd3\n', ['345678'], '\n')]
        expected = [
            (None, 'd1\n', [], ''),
            ('1.0.1', '\nd2\n', ['234567'], '\n'),
            ('999.9.9', '\nd3\n', ['345678'], '\n')]
        self.assertItemsEqual(
            extend_releases(releases, '1.0.1', ['234567']), expected)

    def test_extend_releases_append(self):
        """Test whether the rel. version is appended to the beginning.

        :raises exceptions.AssertionError: if the test fails

        """
        releases = [
            (None, 'd1\n', [], ''),
            ('1.0.1', '\nd2\n', ['234567'], '\n')]
        expected = [
            (None, 'd1\n', [], ''),
            ('999.9.9', '\n', ['345678'], '\n'),
            ('1.0.1', '\nd2\n', ['234567'], '\n')]
        self.assertItemsEqual(
            extend_releases(releases, '999.9.9', ['345678']), expected)

    def test_format_release_version(self):
        """Test whether correct string is returned.

        :raises exceptions.AssertionError: if the test fails

        """
        self.assertEqual(
            format_release('1.0.1', '\ndesc\n', ['123456', '234567'], '\ne\n'),
            '===================\n'
            '1.0.1 Release Notes\n'
            '===================\n'
            '\n'
            'desc\n'
            '* :rhbug:`123456`\n'
            '* :rhbug:`234567`\n'
            '\n'
            'e\n')

    def test_format_release_none(self):
        """Test whether no version is handled properly.

        :raises exceptions.AssertionError: if the test fails

        """
        self.assertEqual(format_release(None, 'l1\nl2\n', [], ''), 'l1\nl2\n')

    def test_update_notes_append(self):
        """Test whether the rel. version is appended to the beginning.

        :raises exceptions.AssertionError: if the test fails

        """
        regex = re.compile(
            '^Update finished. See the diff to the Git index:\n.+')
        notesfn = os.path.join(self.repository.working_dir, NOTES_FN)
        title = TITLE_PAT.format(version=self.VERSION)
        extra = [
            '\n',
            '{}\n'.format('=' * (len(title) - 1)),
            title,
            '{}\n'.format('=' * (len(title) - 1)),
            '\n',
            ISSUE_PAT.format(number=self.ISSUE)]
        with self._read_copy(notesfn) as original:
            # Insert the extra lines right after the 22nd line.
            expected = itertools.chain(
                itertools.islice(original, 0, 22), extra, original)
            with self._patch_file(), self._assert_logs(logging.INFO, regex):
                update_notes()
            with open(notesfn, 'rt') as actual:
                self._assert_iter_equal(actual, expected)

    def test_update_notes_released(self):
        """Test whether the released version is detected.

        :raises exceptions.AssertionError: if the test fails

        """
        self.repository.create_tag(TAG_PAT.format(version=self.VERSION))
        with self._patch_file():
            self.assertRaises(ValueError, update_notes)

    def test_main_append(self):
        """Test whether the release version is appended to the end.

        :raises exceptions.AssertionError: if the test fails

        """
        regex = re.compile(
            '^INFO Update finished. See the diff to the Git index:\n.+', re.M)
        notesfn = os.path.join(self.repository.working_dir, NOTES_FN)
        title = TITLE_PAT.format(version=self.VERSION)
        extra = [
            '\n',
            '{}\n'.format('=' * (len(title) - 1)),
            title,
            '{}\n'.format('=' * (len(title) - 1)),
            '\n',
            ISSUE_PAT.format(number=self.ISSUE)]
        with self._read_copy(notesfn) as original:
            # Insert the extra lines right after the 22nd line.
            expected = itertools.chain(
                itertools.islice(original, 0, 22), extra, original)
            with \
                    tests.mock.patch('sys.argv', ['prog']), \
                    self._patch_file(), \
                    self._assert_prints(regex, 'sys.stderr'):
                main()
            with open(notesfn, 'rt') as actual:
                self._assert_iter_equal(actual, expected)

    def test_main_released(self):
        """Test whether the released version is detected.

        :raises exceptions.AssertionError: if the test fails

        """
        self.repository.create_tag(TAG_PAT.format(version=self.VERSION))
        with \
                tests.mock.patch('sys.argv', ['prog']), self._patch_file(), \
                self.assertRaises(SystemExit) as context:
            main()
        self.assertNotEqual(context.exception.code, 0)

    def test_releasenotesparser(self):
        """Test whether correct release notes are yielded.

        :raises exceptions.AssertionError: if the test fails

        """
        notesfn = os.path.join(self.repository.working_dir, NOTES_FN)
        parser = ReleaseNotesParser()
        descriptions = [
            # None
            '..\n'
            '  Copyright (C) 2014  Red Hat, Inc.\n'
            '\n'
            '  This copyrighted material is made available to anyone wishing '
            'to use,\n  modify, copy, or redistribute it subject to the terms '
            'and conditions of\n  the GNU General Public License v.2, or (at '
            'your option) any later version.\n  This program is distributed in'
            ' the hope that it will be useful, but WITHOUT\n  ANY WARRANTY '
            'expressed or implied, including the implied warranties of\n  '
            'MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU'
            ' General\n  Public License for more details.  You should have '
            'received a copy of the\n  GNU General Public License along with '
            'this program; if not, write to the\n  Free Software Foundation, '
            'Inc., 51 Franklin Street, Fifth Floor, Boston, MA\n  02110-1301, '
            'USA.  Any Red Hat trademarks that are incorporated in the\n  '
            'source code or documentation are not subject to the GNU General '
            'Public\n  License and may only be used or replicated with the '
            'express permission of\n  Red Hat, Inc.\n'
            '\n'
            '###################\n'
            ' DNF Release Notes\n'
            '###################\n'
            '\n'
            '.. contents::\n'
            '\n',
            # 0.3.1
            '\n0.3.1 brings mainly changes to the automatic metadata '
            'synchronization. In\nFedora, ``dnf makecache`` is triggered via '
            'SystemD timers now and takes an\noptional ``background`` '
            'extra-argument to run in resource-considerate mode (no\nsyncing '
            'when running on laptop battery, only actually performing the '
            'check at\nmost once every three hours). Also, the IO and CPU '
            'priorities of the\ntimer-triggered process are lowered now and '
            "shouldn't as noticeably impact the\nsystem's performance.\n\nThe "
            'administrator can also easily disable the automatic metadata '
            'updates by\nsetting :ref:`metadata_timer_sync '
            '<metadata_timer_sync-label>` to 0.\n\nThe default value of '
            ':ref:`metadata_expire <metadata_expire-label>` was\nincreased '
            'from 6 hours to 48 hours. In Fedora, the repos usually set this\n'
            'explicitly so this change is not going to cause much impact.\n\n'
            'The following reported issues are fixed in this release:\n\n',
            # 0.3.2
            '\nThe major improvement in this version is in speeding up syncing'
            ' of repositories\nusing metalink by looking at the repomd.xml '
            'checksums. This effectively lets DNF\ncheaply refresh expired '
            'repositories in cases where the original has not\nchanged\\: for '
            'instance the main Fedora repository is refreshed with one 30 kB\n'
            'HTTP download. This functionality is present in the current Yum '
            "but hasn't\nworked in DNF since 3.0.0.\n\nOtherwise this is "
            'mainly a release fixing bugs and tracebacks. The following\n'
            'reported bugs are fixed:\n\n',
            # 0.3.3
            '\nThe improvements in 0.3.3 are only API changes to the logging. '
            'There is a new\nmodule ``dnf.logging`` that defines simplified '
            'logging structure compared to\nYum, with fewer logging levels and'
            ' `simpler usage for the developers\n<https://github.com/'
            'rpm-software-management/dnf/wiki/Hacking#logging>`_. The RPM '
            'transaction logs are\nno longer in ``/var/log/dnf.transaction.log'
            '`` but in ``/var/log/dnf.rpm.log`` by\ndefault.\n\nThe exception '
            'classes were simplified and moved to ``dnf.exceptions``.\n\nThe '
            'following bugs are fixed in 0.3.3:\n\n',
            # 0.3.4
            '\n0.3.4 is the first DNF version since the fork from Yum that is '
            'able to\nmanipulate the comps data. In practice, ``dnf group '
            'install <group name>`` works\nagain. No other group commands are '
            'supported yet.\n\nSupport for ``librepo-0.0.4`` and related '
            'cleanups and extensions this new\nversion allows are included '
            '(see the buglist below)\n\nThis version has also improved '
            'reporting of obsoleted packages in the CLI (the\nYum-style '
            '"replacing <package-nevra>" appears in the textual transaction\n'
            'overview).\n\nThe following bugfixes are included in 0.3.4:\n\n',
            # 0.3.5
            '\nBesides few fixed bugs this version should not present any '
            'differences for the\nuser. On the inside, the transaction '
            'managing mechanisms have changed\ndrastically, bringing code '
            'simplification, better maintainability and better\ntestability.\n'
            '\nIn Fedora, there is a change in the spec file effectively '
            'preventing the\nmakecache timer from running *immediately after '
            'installation*. The timer\nservice is still enabled by default, '
            'but unless the user starts it manually with\n``systemctl start '
            'dnf-makecache.timer`` it will not run until after the first\n'
            'reboot. This is in alignment with Fedora packaging best '
            'practices.\n\nThe following bugfixes are included in 0.3.5:\n\n',
            # 0.3.6
            '\nThis is a bugfix release, including the following fixes:\n\n',
            # 0.3.7
            '\nThis is a bugfix release:\n\n',
            # 0.3.8
            '\nA new locking module has been integrated in this version, '
            'clients should see the\nmessage about DNF lock being taken less '
            'often.\n\nPanu Matilainen has submitted many patches to this '
            'release to cleanup the RPM\ninterfacing modules.\n\nThe following'
            ' bugs are fixed in this release:\n\n',
            # 0.3.9
            '\nThis is a quick bugfix release dealing with reported bugs and '
            'tracebacks:\n\n',
            # 0.3.10
            '\nThe only major change is that ``skip_if_unavailable`` is '
            ':ref:`enabled by\ndefault now <skip_if_unavailable_default>`.\n\n'
            'A minor release otherwise, mainly to get a new version of DNF out'
            ' that uses a\nfresh librepo. The following issues are now a thing'
            ' of the past:\n\n',
            # 0.3.11
            '\nThe default multilib policy configuration value is ``best`` '
            'now. This does not\npose any change for the Fedora users because '
            'exactly the same default had been\npreviously achieved by a '
            'setting in ``/etc/dnf/dnf.conf`` shipped with the\nFedora '
            'package.\n\nAn important fix to the repo module speeds up package'
            ' downloads again is present\nin this release. The full list of '
            'fixes is:\n\n',
            # 0.4.0
            '\nThe new minor version brings many internal changes to the comps'
            ' code, most comps\nparsing and processing is now delegated to '
            '`libcomps\n<https://github.com/midnightercz/libcomps>`_ by '
            'Jind\xc5\x99ich Lu\xc5\xbea.\n\nThe ``overwrite_groups`` config '
            'option has been dropped in this version and DNF\nacts if it was '
            '0, that is groups with the same name are merged together.\n\nThe '
            'currently supported groups commands (``group list`` and '
            '``group install``)\nare documented on the manpage now.\n\nThe '
            '0.4.0 version is the first one supported by the DNF Payload for '
            'Anaconda and\nmany changes since 0.3.11 make that possible by '
            'cleaning up the API and making\nit more sane (cleanup of '
            '``yumvars`` initialization API, unifying the RPM\ntransaction '
            'callback objects hierarchy, slimming down ``dnf.rpmUtils.arch``,'
            '\nimproved logging).\n\nFixes for the following are contained in '
            'this version:\n\n',
            # 0.4.1
            '\nThe focus of this release was to support our efforts in '
            'implementing the DNF\nPayload for Anaconda, with changes on the '
            'API side of things (better logging,\nnew ``Base.reset()`` '
            'method).\n\nSupport for some irrelevant config options has been '
            'dropped (``kernelpkgnames``,\n``exactarch``, '
            '``rpm_check_debug``). We also no longer detect metalinks in the\n'
            '``mirrorlist`` option (see `Fedora bug 948788\n'
            '<https://bugzilla.redhat.com/show_bug.cgi?id=948788>`_).\n\nDNF '
            'is on its way to drop the urlgrabber dependency and the first set'
            ' of patches\ntowards this goal is already in.\n\nExpect the '
            'following bugs to go away with upgrade to 0.4.1:\n\n',
            # 0.4.2
            '\nDNF now downloads packages for the transaction in parallel with'
            ' progress bars\nupdated to effectively represent this. Since so '
            'many things in the downloading\ncode were changing, we figured it'
            ' was a good idea to finally drop urlgrabber\ndependency at the '
            "same time. Indeed, this is the first version that doesn't\n"
            'require urlgrabber for neither build nor run.\n\nSimilarly, since'
            ' `librepo started to support this\n<https://github.com/Tojaj/'
            'librepo/commit/acf458f29f7234d2d8d93a68391334343beae4b9>`_,\n'
            'downloads in DNF now use the fastests mirrors available by '
            "default.\n\nThe option to :ref:`specify repositories' costs "
            '<repo_cost-label>` has been\nreadded.\n\nInternally, DNF has seen'
            ' first part of ongoing refactorings of the basic\noperations '
            '(install, update) as well as a couple of new API methods '
            'supporting\ndevelopment of extensions.\n\nThese bugzillas are '
            'fixed in 0.4.2:\n\n',
            # 0.4.3
            '\nThis is an early release to get the latest DNF out with the '
            'latest librepo\nfixing the `Too many open files\n'
            '<https://bugzilla.redhat.com/show_bug.cgi?id=1015957>`_ bug.\n\n'
            'In Fedora, the spec file has been updated to no longer depend on '
            'precise\nversions of the libraries so in the future they can be '
            'released\nindependently.\n\nThis release sees the finished '
            'refactoring in error handling during basic\noperations and adds '
            'support for ``group remove`` and ``group info`` commands,\ni.e. '
            'the following two bugs:\n\n',
            # 0.4.4
            '\nThe initial support for Python 3 in DNF has been merged in this'
            ' version. In\npractice one can not yet run the ``dnf`` command in'
            ' Py3 but the unit tests\nalready pass there. We expect to give '
            'Py3 and DNF heavy testing during the\nFedora 21 development cycle'
            ' and eventually switch to it as the default. The plan\nis to drop'
            ' Python 2 support as soon as Anaconda is running in Python 3.\n\n'
            'Minor adjustments to allow Anaconda support also happened during '
            'the last week,\nas well as a fix to a possibly severe bug that '
            'one is however not really likely\nto see with non-devel Fedora '
            'repos:\n\n',
            # 0.4.5
            '\nA serious bug causing `tracebacks during package downloads\n'
            '<https://bugzilla.redhat.com/show_bug.cgi?id=1021087>`_ made it '
            'into 0.4.4 and\nthis release contains a fix for that. Also, a '
            'basic proxy support has been\nreadded now.\n\nBugs fixed in '
            '0.4.5:\n\n',
            # 0.4.6
            '\n0.4.6 brings two new major features. Firstly, it is the revival'
            ' of ``history\nundo``, so transactions can be reverted now.  '
            'Secondly, DNF will now limit the\nnumber of installed kernels and'
            ' *installonly* packages in general to the number\nspecified by '
            ':ref:`installonly_limit <installonly-limit-label>` configuration'
            '\noption.\n\nDNF now supports the ``group summary`` command and '
            'one-word group commands no\nlonger cause tracebacks, e.g. '
            '``dnf grouplist``.\n\nThere are vast internal changes to '
            '``dnf.cli``, the subpackage that provides CLI\nto DNF. In '
            'particular, it is now better separated from the core.\n\nThe '
            'hawkey library used against DNF from with this versions uses a '
            '`recent RPMDB\nloading optimization in libsolv\n'
            '<https://github.com/openSUSE/libsolv/commit/843dc7e1>`_ that '
            'shortens DNF\nstartup by seconds when the cached RPMDB is '
            'invalid.\n\nWe have also added further fixes to support Python 3 '
            "and enabled `librepo's\nfastestmirror caching optimization\n"
            '<https://github.com/Tojaj/librepo/commit/'
            'b8a063763ccd8a84b8ec21a643461eaace9b9c08>`_\nto tighten the '
            'download times even more.\n\nBugs fixed in 0.4.6:\n\n',
            # 0.4.7
            '\nWe start to publish the :doc:`api` with this release. It is '
            'largely\nincomprehensive at the moment, yet outlines the shape of'
            ' the documentation and\nthe process the project is going to use '
            'to maintain it.\n\nThere are two Yum configuration options that '
            'were dropped: :ref:`group_package_types '
            '<group_package_types_dropped>` and '
            ':ref:`upgrade_requirements_on_install '
            '<upgrade_requirements_on_install_dropped>`.\n\nBugs fixed in '
            '0.4.7:\n\n',
            # 0.4.8
            '\nThere are mainly internal changes, new API functions and '
            'bugfixes in this release.\n\nPython 3 is fully supported now, the'
            ' Fedora builds include the Py3 variant. The DNF program still '
            'runs under Python 2.7 but the extension authors can now choose '
            'what Python they prefer to use.\n\nThis is the first version of '
            'DNF that deprecates some of its API. Clients using deprecated '
            'code will see a message emitted to stderr using the standard '
            '`Python warnings module '
            '<http://docs.python.org/3.3/library/warnings.html>`_. You can '
            'filter out :exc:`dnf.exceptions.DeprecationWarning` to suppress '
            'them.\n\nAPI additions in 0.4.8:\n\n* :attr:`dnf.Base.sack`\n* '
            ':attr:`dnf.conf.Conf.cachedir`\n* '
            ':attr:`dnf.conf.Conf.config_file_path`\n* '
            ':attr:`dnf.conf.Conf.persistdir`\n* :meth:`dnf.conf.Conf.read`\n*'
            ' :class:`dnf.package.Package`\n* :class:`dnf.query.Query`\n* '
            ':class:`dnf.subject.Subject`\n* :meth:`dnf.repo.Repo.__init__`\n*'
            ' :class:`dnf.sack.Sack`\n* :class:`dnf.selector.Selector`\n* '
            ':class:`dnf.transaction.Transaction`\n\nAPI deprecations in '
            '0.4.8:\n\n* :mod:`dnf.queries` is deprecated now. If you need to '
            'create instances of :class:`.Subject`, import it from '
            ':mod:`dnf.subject`. To create :class:`.Query` instances it is '
            'recommended to use :meth:`sack.query() <dnf.sack.Sack.query>`.\n'
            '\nBugs fixed in 0.4.8:\n\n',
            # 0.4.9
            '\nSeveral Yum features are revived in this release. '
            '``dnf history rollback`` now works again. The '
            '``history userinstalled`` has been added, it displays a list of '
            'ackages that the user manually selected for installation on an '
            'installed system and does not include those packages that got '
            "installed as dependencies.\n\nWe're happy to announce that the "
            'API in 0.4.9 has been extended to finally support plugins. There '
            'is a limited set of plugin hooks now, we will carefully add new '
            'ones in the following releases. New marking operations have ben '
            'added to the API and also some configuration options.\n\nAn '
            'alternative to ``yum shell`` is provided now for its most common '
            'use case: :ref:`replacing a non-leaf package with a conflicting '
            'package <allowerasing_instead_of_shell>` is achieved by using the'
            ' ``--allowerasing`` switch now.\n\nAPI additions in 0.4.9:\n\n* '
            ':doc:`api_plugins`\n* :ref:`logging_label`\n* '
            ':meth:`.Base.read_all_repos`\n* :meth:`.Base.reset`\n* '
            ':meth:`.Base.downgrade`\n* :meth:`.Base.remove`\n* '
            ':meth:`.Base.upgrade`\n* :meth:`.Base.upgrade_all`\n* '
            ':attr:`.Conf.pluginpath`\n* :attr:`.Conf.reposdir`\n\nAPI '
            'deprecations in 0.4.9:\n\n* :exc:`.PackageNotFoundError` is '
            'deprecated for public use. Please catch :exc:`.MarkingError` '
            'instead.\n* It is deprecated to use :meth:`.Base.install` return '
            'value for anything. The method either returns or raises an '
            'exception.\n\nBugs fixed in 0.4.9:\n\n',
            # 0.4.10
            '\n0.4.10 is a bugfix release that also adds some long-requested '
            'CLI features and extends the plugin support with two new plugin '
            'hooks. An important feature for plugin developers is going to be '
            "the possibility to register plugin's own CLI command, available "
            'from this version.\n\n``dnf history`` now recognizes ``last`` as '
            'a special argument, just like other history commands.\n\n'
            '``dnf install`` now accepts group specifications via the ``@`` '
            'character.\n\nSupport for the ``--setopt`` option has been '
            'readded from Yum.\n\nAPI additions in 0.4.10:\n\n* '
            ':doc:`api_cli`\n* :attr:`.Plugin.name`\n* '
            ':meth:`.Plugin.__init__` now specifies the second parameter as an'
            ' instance of `.cli.Cli`\n* :meth:`.Plugin.sack`\n* '
            ':meth:`.Plugin.transaction`\n* :func:`.repo.repo_id_invalid`\n\n'
            'API changes in 0.4.10:\n\n* Plugin authors must specify '
            ':attr:`.Plugin.name` when authoring a plugin.\n\nBugs fixed in '
            '0.4.10:\n\n',
            # 0.4.11
            '\nThis is mostly a bugfix release following quickly after 0.4.10,'
            ' with many updates to documentation.\n\nAPI additions in 0.4.11:'
            '\n\n* :meth:`.Plugin.read_config`\n* :class:`.repo.Metadata`\n* '
            ':attr:`.repo.Repo.metadata`\n\nAPI changes in 0.4.11:\n\n* '
            ':attr:`.Conf.pluginpath` is no longer hard coded but depends on '
            'the major Python version.\n\nBugs fixed in 0.4.11:\n\n',
            # 0.4.12
            '\nThis release disables fastestmirror by default as we received '
            'many complains about it. There are also several bugfixes, most '
            'importantly an issue has been fixed that caused packages '
            'installed by Anaconda be removed together with a depending '
            'package. It is now possible to use ``bandwidth`` and ``throttle``'
            ' config values too.\n\nBugs fixed in 0.4.12:\n\n',
            # 0.4.13
            '\n0.4.13 finally ships support for `delta RPMS '
            '<https://gitorious.org/deltarpm>`_. Enabling this can save some '
            'bandwidth (and use some CPU time) when downloading packages for '
            'updates.\n\nSupport for bash completion is also included in this '
            'version. It is recommended to use the '
            '``generate_completion_cache`` plugin to have the completion work '
            'fast. This plugin will be also shipped with '
            '``dnf-plugins-core-0.0.3``.\n\nThe '
            ':ref:`keepcache <keepcache-label>` config option has been '
            'readded.\n\nBugs fixed in 0.4.13:\n\n',
            # 0.4.14
            '\nThis quickly follows 0.4.13 to address the issue of crashes '
            'when DNF output is piped into another program.\n\nAPI additions '
            'in 0.4.14:\n\n* :attr:`.Repo.pkgdir`\n\nBugs fixed in 0.4.14:\n'
            '\n',
            # 0.4.15
            '\nMassive refactoring of the downloads handling to provide better'
            ' API for reporting download progress and fixed bugs are the main '
            'things brought in 0.4.15.\n\nAPI additions in 0.4.15:\n\n* '
            ':exc:`dnf.exceptions.DownloadError`\n* '
            ':meth:`dnf.Base.download_packages` now takes the optional '
            '`progress` parameter and can raise :exc:`.DownloadError`.\n* '
            ':class:`dnf.callback.Payload`\n* '
            ':class:`dnf.callback.DownloadProgress`\n* '
            ':meth:`dnf.query.Query.filter` now also recognizes ``provides`` '
            'as a filter name.\n\nBugs fixed in 0.4.15:\n\n',
            # 0.4.16
            '\nThe refactorings from 0.4.15 are introducing breakage causing '
            'the background ``dnf makecache`` runs traceback. This release '
            'fixes that.\n\nBugs fixed in 0.4.16:\n\n',
            # 0.4.17
            '\nThis release fixes many bugs in the downloads/DRPM CLI area. A '
            'bug got fixed preventing a regular user from running read-only '
            'operations using ``--cacheonly``. Another fix ensures that '
            '``metadata_expire=never`` setting is respected. Lastly, the '
            'release provides three requested API calls in the repo management'
            ' area.\n\nAPI additions in 0.4.17:\n\n* '
            ':meth:`dnf.repodict.RepoDict.all`\n* '
            ':meth:`dnf.repodict.RepoDict.get_matching`\n* '
            ':meth:`dnf.repo.Repo.set_progress_bar`\n\nBugs fixed in 0.4.17:\n'
            '\n',
            # 0.4.18
            '\nSupport for ``dnf distro-sync <spec>`` finally arrives in this '
            'version.\n\nDNF has moved to handling groups as objects,  tagged '
            'installed/uninstalled independently from the actual installed '
            'packages. This has been in Yum as the ``group_command=objects`` '
            'setting and the default in recent Fedora releases. There are API '
            'extensions related to this change as well as two new CLI '
            'commands: ``group mark install`` and ``group mark remove``.\n\n'
            'API items deprecated in 0.4.8 and 0.4.9 have been dropped in '
            '0.4.18, in accordance with our :ref:`deprecating-label`.\n\nAPI '
            'changes in 0.4.18:\n\n* :mod:`dnf.queries` has been dropped as '
            'announced in `0.4.8 Release Notes`_\n* '
            ':exc:`dnf.exceptions.PackageNotFoundError` has been dropped from '
            'API as announced in `0.4.9 Release Notes`_\n* '
            ':meth:`dnf.Base.install` no longer has to return the number of '
            'marked packages as announced in `0.4.9 Release Notes`_\n\nAPI '
            'deprecations in 0.4.18:\n\n* :meth:`dnf.Base.select_group` is '
            'deprecated now. Please use :meth:`~.Base.group_install` instead.'
            '\n\nAPI additions in 0.4.18:\n\n* :meth:`dnf.Base.group_install`'
            '\n* :meth:`dnf.Base.group_remove`\n\nBugs fixed in 0.4.18:\n\n',
            # 0.4.19
            '\nArriving one week after 0.4.18, the 0.4.19 mainly provides a '
            'fix to a traceback in group operations under non-root users.\n\n'
            'DNF starts to ship separate translation files (.mo) starting with'
            ' this release.\n\nBugs fixed in 0.4.19:\n\n',
            # 0.5.0
            '\nThe biggest improvement in 0.5.0 is complete support for groups'
            ' `and environments '
            '<https://bugzilla.redhat.com/show_bug.cgi?id=1063666>`_, '
            'including internal database of installed groups independent of '
            'the actual packages (concept known as groups-as-objects from '
            'Yum). Upgrading groups is supported now with ``group upgrade`` '
            'too.\n\nTo force refreshing of metadata before an operation (even'
            ' if the data is not expired yet), `the refresh option has been '
            'added <https://bugzilla.redhat.com/show_bug.cgi?id=1064226>`_.\n'
            '\nInternally, the CLI went through several changes to allow for '
            'better API accessibility like `granular requesting of root '
            'permissions '
            '<https://bugzilla.redhat.com/show_bug.cgi?id=1062889>`_.\n\nAPI '
            'has got many more extensions, focusing on better manipulation '
            'with comps and packages. There are new entries in '
            ':doc:`cli_vs_yum` and :doc:`user_faq` too.\n\nSeveral resource '
            'leaks (file descriptors, noncollectable Python objects) were '
            'found and fixed.\n\nAPI changes in 0.5.0:\n\n* it is now '
            'recommended that either :meth:`dnf.Base.close` is used, or that '
            ':class:`dnf.Base` instances are treated as a context manager.\n\n'
            'API extensions in 0.5.0:\n\n* :meth:`dnf.Base.add_remote_rpms`\n* '
            ':meth:`dnf.Base.close`\n* :meth:`dnf.Base.group_upgrade`\n* '
            ':meth:`dnf.Base.resolve` optionally accepts `allow_erasing` '
            'arguments now.\n* :meth:`dnf.Base.package_downgrade`\n* '
            ':meth:`dnf.Base.package_install`\n* '
            ':meth:`dnf.Base.package_upgrade`\n* '
            ':class:`dnf.cli.demand.DemandSheet`\n* '
            ':attr:`dnf.cli.Command.base`\n* :attr:`dnf.cli.Command.cli`\n* '
            ':attr:`dnf.cli.Command.summary`\n* :attr:`dnf.cli.Command.usage`'
            '\n* :meth:`dnf.cli.Command.configure`\n* '
            ':attr:`dnf.cli.Cli.demands`\n* :class:`dnf.comps.Package`\n* '
            ':meth:`dnf.comps.Group.packages_iter`\n* '
            ':data:`dnf.comps.MANDATORY` etc.\n\nBugs fixed in 0.5.0:\n\n',
            # 0.5.1
            '\nBugfix release with several internal cleanups. One outstanding '
            'change for CLI users is that DNF is a lot less verbose now during'
            ' the dependency resolving phase.\n\nBugs fixed in 0.5.1:\n\n',
            # 0.5.2
            '\nThis release brings `autoremove command '
            '<https://bugzilla.redhat.com/show_bug.cgi?id=963345>`_ that '
            'removes any package that was originally installed as a dependency'
            ' (e.g. had not been specified as an explicit argument to the '
            'install command) and is no longer needed.\n\nEnforced '
            'verification of SSL connections can now be disabled with the '
            ':ref:`sslverify setting <sslverify-label>`.\n\nWe have been '
            'plagued with many crashes related to Unicode and encodings since '
            "the 0.5.0 release. These have been cleared out now.\n\nThere's "
            'more: improvement in startup time, `extended globbing semantics '
            'for input arguments '
            '<https://bugzilla.redhat.com/show_bug.cgi?id=1083679>`_ and '
            '`better search relevance sorting '
            '<https://bugzilla.redhat.com/show_bug.cgi?id=1093888>`_.\n\nBugs '
            'fixed in 0.5.2:\n\n',
            # 0.5.3
            '\nA set of bugfixes related to i18n and Unicode handling. There '
            'is a ``-4/-6`` switch and a corresponding :ref:`ip_resolve '
            '<ip-resolve-label>` configuration option (both known from Yum) to'
            ' force DNS resolving of hosts to IPv4 or IPv6 addresses.\n\n0.5.3'
            ' comes with several extensions and clarifications in the API: '
            'notably :class:`~.dnf.transaction.Transaction` is introspectible '
            'now, :class:`Query.filter <dnf.query.Query.filter>` is more '
            "useful with new types of arguments and we've hopefully shed more"
            ' light on how a client is expected to setup the configuration '
            ':attr:`~dnf.conf.Conf.substitutions`.\n\nFinally, plugin authors '
            'can now use a new :meth:`~dnf.Plugin.resolved` hook.\n\nAPI '
            'changes in 0.5.3:\n\n* extended description given for '
            ':meth:`dnf.Base.fill_sack`\n* :meth:`dnf.Base.select_group` has '
            'been dropped as announced in `0.4.18 Release Notes`_\n\nAPI '
            'additions in 0.5.3:\n\n* :attr:`dnf.conf.Conf.substitutions`\n* '
            ':attr:`dnf.package.Package.arch`\n* '
            ':attr:`dnf.package.Package.buildtime`\n* '
            ':attr:`dnf.package.Package.epoch`\n* '
            ':attr:`dnf.package.Package.installtime`\n* '
            ':attr:`dnf.package.Package.name`\n* '
            ':attr:`dnf.package.Package.release`\n* '
            ':attr:`dnf.package.Package.sourcerpm`\n* '
            ':attr:`dnf.package.Package.version`\n* '
            ':meth:`dnf.Plugin.resolved`\n* :meth:`dnf.query.Query.filter` '
            'accepts suffixes for its argument keys now which change the '
            'filter semantics.\n* :mod:`dnf.rpm`\n* '
            ':class:`dnf.transaction.TransactionItem`\n* '
            ':class:`dnf.transaction.Transaction` is iterable now.\n\nBugs '
            'fixed in 0.5.3:\n\n',
            # 0.5.4
            '\nSeveral encodings bugs were fixed in this release, along with '
            'some packaging issues and updates to :doc:`conf_ref`.\n\n'
            'Repository :ref:`priority <repo_priority-label>` configuration '
            'setting has been added, providing similar functionality to Yum '
            "Utils' Priorities plugin.\n\nBugs fixed in 0.5.4:\n\n",
            # 0.5.5
            '\nThe full proxy configuration, API extensions and several '
            'bugfixes are provided in this release.\n\nAPI changes in 0.5.5:\n'
            '\n* `cachedir`, the second parameter of '
            ':meth:`dnf.repo.Repo.__init__` is not optional (the method has '
            'always been this way but the documentation was not matching)\n\n'
            'API additions in 0.5.5:\n\n* extended description and an example '
            'provided for :meth:`dnf.Base.fill_sack`\n* '
            ':attr:`dnf.conf.Conf.proxy`\n* '
            ':attr:`dnf.conf.Conf.proxy_username`\n* '
            ':attr:`dnf.conf.Conf.proxy_password`\n* '
            ':attr:`dnf.repo.Repo.proxy`\n* '
            ':attr:`dnf.repo.Repo.proxy_username`\n* '
            ':attr:`dnf.repo.Repo.proxy_password`\n\nBugs fixed in 0.5.5:\n\n',
            # 0.6.0
            '\n0.6.0 marks a new minor version of DNF and the first release to'
            ' support advisories listing with the :ref:`udpateinfo command '
            '<updateinfo_command-label>`.\n\nSupport for the :ref:`include '
            'configuration directive <include-label>` has been added. Its '
            "functionality reflects Yum's ``includepkgs`` but it has been "
            'renamed to make it consistent with the ``exclude`` setting.\n\n'
            'Group operations now produce a list of proposed marking changes '
            'to group objects and the user is given a chance to accept or '
            'reject them just like with an ordinary package transaction.\n\n'
            'Bugs fixed in 0.6.0:\n\n',
            # 0.6.1
            '\nNew release adds :ref:`upgrade-type command '
            '<upgrade_type_automatic-label>` to `dnf-automatic` for choosing '
            'specific advisory type updates.\n\nImplemented missing '
            ':ref:`history redo command <history_redo_command-label>` for '
            'repeating transactions.\n\nSupports :ref:`gpgkey '
            '<repo_gpgkey-label>` repo config, :ref:`repo_gpgcheck '
            '<repo_gpgcheck-label>` and :ref:`gpgcheck <gpgcheck-label>` '
            '[main] and Repo configs.\n\nDistributing new package '
            ':ref:`dnf-yum <dnf_yum_package-label>` that provides '
            '`/usr/bin/yum` as a symlink to `/usr/bin/dnf`.\n\nAPI additions '
            'in 0.6.1:\n\n* `exclude`, the third parameter of '
            ':meth:`dnf.Base.group_install` now also accepts glob patterns of '
            'package names.\n\nBugs fixed in 0.6.1:\n\n',
            # 0.6.2
            '\nAPI additions in 0.6.2:\n\n* Now '
            ':meth:`dnf.Base.package_install` method ignores already installed'
            ' packages\n* `CliError` exception from :mod:`dnf.cli` documented'
            '\n* `Autoerase`, `History`, `Info`, `List`, `Provides`, '
            '`Repolist` commands do not force a sync of expired :ref:`metadata'
            ' <metadata_synchronization-label>`\n* `Install` command does '
            'installation only\n\nBugs fixed in 0.6.2:\n\n',
            # 0.6.3
            '\n:ref:`Deltarpm <deltarpm-label>` configuration option is set on'
            ' by default.\n\nAPI additions in 0.6.3:\n\n* dnf-automatic adds '
            ':ref:`motd emitter <emit_via_automatic-label>` as an alternative '
            'output\n\nBugs fixed in 0.6.3:\n\n',
            # 0.6.4
            '\nAdded example code snippets into :doc:`use_cases`.\n\nShows '
            'ordered groups/environments by `display_order` tag from :ref:`cli'
            ' <grouplist_command-label>` and :doc:`api_comps` DNF API.\n\nIn '
            'commands the environment group is specified the same as '
            ':ref:`group <specifying_groups-label>`.\n\n'
            ':ref:`skip_if_unavailable <skip_if_unavailable-label>` '
            'configuration option affects the metadata only.\n\nadded '
            '`enablegroups`, `minrate` and `timeout` :doc:`configuration '
            'options <conf_ref>`\n\nAPI additions in 0.6.4:\n\nDocumented '
            '`install_set` and `remove_set attributes` from '
            ':doc:`api_transaction`.\n\nExposed `downloadsize`, `files`, '
            '`installsize` attributes from :doc:`api_package`.\n\nBugs fixed '
            'in 0.6.4:\n\n',
            # 0.6.5
            '\nPython 3 version of DNF is now default in Fedora 23 and later.'
            '\n\nyum-dnf package does not conflict with yum package.\n\n'
            '`dnf erase` was deprecated in favor of `dnf remove`.\n\nExtended '
            'documentation of handling non-existent packages and YUM to DNF '
            'transition in :doc:`cli_vs_yum`.\n\nAPI additions in 0.6.5:\n\n'
            'Newly added `pluginconfpath` option in :doc:`configuration '
            '<conf_ref>`.\n\nExposed `skip_if_unavailable` attribute from '
            ':doc:`api_repos`.\n\nDocumented `IOError` exception of method '
            '`fill_sack` from :class:`dnf.Base`.\n\nBugs fixed in 0.6.5:\n\n',
            # 1.0.0
            '\nImproved documentation of YUM to DNF transition in '
            ':doc:`cli_vs_yum`.\n\n:ref:`Auto remove command '
            '<autoremove_command-label>` does not remove `installonly` '
            'packages.\n\n:ref:`Downgrade command <downgrade_command-label>` '
            'downgrades to specified package version if that is lower than '
            'currently installed one.\n\nDNF now uses :attr:`dnf.repo.Repo.id`'
            ' as a default value for :attr:`dnf.repo.Repo.name`.\n\nAdded '
            'support of repositories which use basic HTTP authentication.\n\n'
            'API additions in 1.0.0:\n\n:doc:`configuration <conf_ref>` '
            'options `username` and `password` (HTTP authentication)\n\n'
            ':attr:`dnf.repo.Repo.username` and :attr:`dnf.repo.Repo.password`'
            ' (HTTP authentication)\n\nBugs fixed in 1.0.0:\n\n',
            # 1.0.1
            '\nDNF follows the Semantic Versioning as defined at '
            '`<http://semver.org/>`_.\n\nDocumented SSL '
            ':doc:`configuration <conf_ref>` and :doc:`repository <api_repos>`'
            ' options.\n\nAdded virtual provides allowing installation of DNF'
            ' commands by their name in the form of\n'
            '``dnf install dnf-command(name)``.\n\n'
            ':doc:`dnf-automatic <automatic>` now by default waits random '
            'interval between 0 and 300 seconds before any network '
            'communication is performed.\n\n\nBugs fixed in 1.0.1:\n\n'
        ]
        rest = [
            (None, [], ''),
            ('0.3.1',
             ['916657', '921294', '922521', '926871', '878826', '922664',
              '892064', '919769'],
             '\n'),
            ('0.3.2', ['947258', '889202', '923384'], '\n'),
            ('0.3.3', ['950722', '903775'], '\n'),
            ('0.3.4', ['887317', '914919', '922667'], '\n'),
            ('0.3.5', ['958452', '959990', '961549', '962188'], '\n'),
            ('0.3.6',
             ['966372', '965410', '963627', '965114', '964467', '963680',
              '963133'],
             '\n'),
            ('0.3.7', ['916662', '967732'], '\n'),
            ('0.3.8',
             ['908491', '968159', '974427', '974866', '976652', '975858'],
             '\n'),
            ('0.3.9', ['964584', '979942', '980227', '981310'], '\n'),
            ('0.3.10', ['977661', '984483', '986545'], '\n'),
            ('0.3.11', ['979042', '977753', '996138', '993916'], '\n'),
            ('0.4.0', ['997403', '1002508', '1002798'], '\n'),
            ('0.4.1', ['998859', '1006366', '1008444', '1003220'], '\n'),
            ('0.4.2', ['909744', '984529', '967798', '995459'], '\n'),
            ('0.4.3', ['1013764', '1013773'], '\n'),
            ('0.4.4', ['1017278'], '\n'),
            ('0.4.5', ['1021087'], '\n'),
            ('0.4.6',
             ['878348', '880524', '1019957', '1020101', '1020934', '1023486'],
             '\n'),
            ('0.4.7', ['1019170', '1024776', '1025650'], '\n'),
            ('0.4.8',
             ['1014563', '1029948', '1030998', '1030297', '1030980'],
             '\n'),
            ('0.4.9',
             ['884615', '963137', '991038', '1032455', '1034607', '1036116'],
             '\n'),
            ('0.4.10',
             ['967264', '1018284', '1035164', '1036147', '1036211', '1038403',
              '1038937', '1040255', '1044502', '1044981', '1044999'],
             '\n'),
            ('0.4.11',
             ['1048402', '1048572', '1048716', '1048719', '1048988'],
             '\n'),
            ('0.4.12',
             ['1045737', '1048468', '1048488', '1049025', '1051554'],
             '\n'),
            ('0.4.13',
             ['909468', '1030440', '1046244', '1055051', '1056400'],
             '\n'),
            ('0.4.14', ['1062390', '1062847', '1063022', '1064148'],
             '\n'),
            ('0.4.15',
             ['1048788', '1065728', '1065879', '1065959', '1066743'],
             '\n'),
            ('0.4.16', ['1069996'], '\n'),
            ('0.4.17',
             ['1059704', '1058224', '1069538', '1070598', '1070710', '1071323',
              '1071455', '1071501', '1071518', '1071677'],
             '\n'),
            ('0.4.18', ['963710', '1067136', '1071212', '1071501'], '\n'),
            ('0.4.19', ['1077173', '1078832', '1079621'], '\n'),
            ('0.5.0',
             ['1029022', '1051869', '1061780', '1062884', '1062889', '1063666',
              '1064211', '1064226', '1073859', '1076884', '1079519', '1079932',
              '1080331', '1080489', '1082230', '1083432', '1083767', '1084139',
              '1084553', '1088166'],
             '\n'),
            ('0.5.1', ['1065882', '1081753', '1089864'], '\n'),
            ('0.5.2',
             ['963345', '1073457', '1076045', '1083679', '1092006', '1092777',
              '1093888', '1094594', '1095580', '1095861', '1096506'],
             '\n'),
            ('0.5.3',
             ['1047049', '1067156', '1093420', '1104757', '1105009', '1110800',
              '1111569', '1111997', '1112669', '1112704'],
             '\n'),
            ('0.5.4',
             ['1048973', '1108908', '1116544', '1116839', '1116845', '1117102',
              '1117293', '1117678', '1118178', '1118796', '1119032'],
             '\n'),
            ('0.5.5',
             ['1100946', '1117789', '1120583', '1121280', '1122900',
              '1123688'],
             '\n'),
            ('0.6.0',
             ['850912', '1055910', '1116666', '1118272', '1127206'],
             '\n'),
            ('0.6.1',
             ['1132335', '1071854', '1131969', '908764', '1130878', '1130432',
              '1118236', '1109915'],
             '\n'),
            ('0.6.2',
             ['909856', '1134893', '1138700', '1070902', '1124316', '1136584',
              '1135861', '1136223', '1122617', '1133830', '1121184'],
             '\n'),
            ('0.6.3',
             ['1153543', '1151231', '1163063', '1151854', '1151740', '1110780',
              '1149972', '1150474', '995537', '1149952', '1149350', '1170232',
              '1147523', '1148208', '1109927'],
             '\n'),
            ('0.6.4',
             ['1155877', '1175466', '1175466', '1186461', '1170156', '1184943',
              '1177002', '1169165', '1167982', '1157233', '1138096', '1181189',
              '1181397', '1175434', '1162887', '1156084', '1175098', '1174136',
              '1055910', '1155918', '1119030', '1177394', '1154476'],
             '\n'),
            ('0.6.5',
             ['1203151', '1187579', '1185977', '1195240', '1193914', '1195385',
              '1160806', '1186710', '1207726', '1157233', '1190671', '1191579',
              '1195325', '1154202', '1189083', '1193915', '1195661', '1190458',
              '1194685', '1160950'],
             '\n'),
            ('1.0.0',
             ['1215560', '1199648', '1208773', '1208018', '1207861', '1201445',
              '1210275', '1191275', '1207965', '1215289'],
             '\n'),
            ('1.0.1',
             ['1214968', '1222694', '1225246', '1213985', '1225277', '1223932',
              '1223614', '1203661', '1187741'],
             '')]
        expected = (
            (version, desc, issues, epilog)
            for desc, (version, issues, epilog) in zip(descriptions, rest))
        with open(notesfn) as notesfile:
            self.assertItemsEqual(parser.parse_lines(notesfile), expected)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# Copyright (C) 2012-2018 Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#

from __future__ import absolute_import
from __future__ import unicode_literals

import operator

import dnf.util

import tests.support
from tests.support import mock


class Slow(object):
    def __init__(self, val):
        self._val = val
        self.computed = 0

    def new_val(self, val):
        self._val = val
        del self._square1
        del self._square2

    @dnf.util.lazyattr("_square1")
    def square1(self):
        self.computed += 1
        return self._val * self._val

    @property
    @dnf.util.lazyattr("_square2")
    def square2(self):
        self.computed += 1
        return self._val * self._val


class Util(tests.support.TestCase):
    def test_am_i_root(self):
        with mock.patch('os.geteuid', return_value=1001):
            self.assertFalse(dnf.util.am_i_root())
        with mock.patch('os.geteuid', return_value=0):
            assert(dnf.util.am_i_root())

    def test_bunch(self):
        b = dnf.util.Bunch()
        self.assertRaises(AttributeError, lambda: b.more)
        b.garden = 'weeds'
        self.assertEqual(b['garden'], 'weeds')
        b['digging'] = 4
        self.assertEqual(b.digging, 4)

    def test_empty(self):
        self.assertTrue(dnf.util.empty(()))
        self.assertFalse(dnf.util.empty([1, 2, 3]))
        self.assertTrue(dnf.util.empty((x for x in [])))
        self.assertTrue(dnf.util.empty(iter([])))
        self.assertFalse(dnf.util.empty((x for x in [2, 3])))

    def test_file_timestamp(self):
        stat = mock.Mock()
        stat.st_mtime = 123
        with mock.patch('os.stat', return_value=stat):
            self.assertEqual(dnf.util.file_timestamp("/yeah"), 123)
        self.assertRaises(OSError, dnf.util.file_timestamp, "/does/not/ex1st")

    def test_first(self):
        self.assertEqual(dnf.util.first([5, 4, 3]), 5)
        ge = (x for x in range(5, 8))
        self.assertEqual(dnf.util.first(ge), 5)
        self.assertEqual(dnf.util.first([]), None)

        def generator():
            if False:
                yield 10
        self.assertEqual(dnf.util.first(generator()), None)

    def test_get_in(self):
        dct = {1: {2: 3},
               5: {8: {9: 10}}}

        self.assertEqual(dnf.util.get_in(dct, (5, 8, 9), -3), 10)
        self.assertEqual(dnf.util.get_in(dct, (5, 8, 8), -3), -3)
        self.assertEqual(dnf.util.get_in(dct, (0, 8, 8), -3), -3)

    def test_group_by_filter(self):
        self.assertEqual(dnf.util.group_by_filter(lambda x: x % 2, range(5)),
                         ([1, 3], [0, 2, 4]))
        self.assertEqual(dnf.util.group_by_filter(lambda x: x, range(5)),
                         ([1, 2, 3, 4], [0]))

    def test_insert_if(self):
        """Test insert_if with sometimes fulfilled condition."""
        item = object()
        iterable = range(4)

        def condition(item):
            return item % 2 == 0

        iterator = dnf.util.insert_if(item, iterable, condition)

        self.assertEqual(next(iterator), item)
        self.assertEqual(next(iterator), 0)
        self.assertEqual(next(iterator), 1)
        self.assertEqual(next(iterator), item)
        self.assertEqual(next(iterator), 2)
        self.assertEqual(next(iterator), 3)
        self.assertRaises(StopIteration, next, iterator)

    def test_is_exhausted_true(self):
        """Test is_exhausted with an iterator which is exhausted."""
        iterator = iter(())

        result = dnf.util.is_exhausted(iterator)

        self.assertTrue(result)

    def test_is_exhausted_false(self):
        """Test is_exhausted with an iterator which is not exhausted."""
        iterator = iter((1,))

        result = dnf.util.is_exhausted(iterator)

        self.assertFalse(result)

    def test_is_glob_pattern(self):
        assert(dnf.util.is_glob_pattern("all*.ext"))
        assert(dnf.util.is_glob_pattern("all?.ext"))
        assert(not dnf.util.is_glob_pattern("not.ext"))

    def test_lazyattr(self):
        slow = Slow(12)

        self.assertEqual(slow.computed, 0)
        self.assertEqual(slow.square1(), 144)
        self.assertEqual(slow.computed, 1)
        self.assertEqual(slow.square1(), 144)
        self.assertEqual(slow.square1(), 144)
        self.assertEqual(slow.computed, 1)

        self.assertEqual(slow.square2, 144)
        self.assertEqual(slow.computed, 2)
        self.assertEqual(slow.square2, 144)
        self.assertEqual(slow.computed, 2)

        slow.new_val(13)
        self.assertEqual(slow.square1(), 169)
        self.assertEqual(slow.square2, 169)
        self.assertEqual(slow.computed, 4)

    def test_mapall(self):
        l = [1, 2, 3]
        out = dnf.util.mapall(lambda n: 2 * n, l)
        self.assertIsInstance(out, list)
        self.assertEqual(out, [2, 4, 6])

    def test_partition(self):
        l = list(range(6))
        smaller, larger = dnf.util.partition(lambda i: i > 4, l)
        self.assertCountEqual(smaller, (0, 1, 2, 3, 4))
        self.assertCountEqual(larger, (5,))

    def test_split_by(self):
        """Test split_by with sometimes fulfilled condition."""
        iterable = range(7)

        def condition(item):
            return item % 3 == 0

        iterator = dnf.util.split_by(iterable, condition)

        self.assertEqual(next(iterator), ())
        self.assertEqual(next(iterator), (0, 1, 2))
        self.assertEqual(next(iterator), (3, 4, 5))
        self.assertEqual(next(iterator), (6,))
        self.assertRaises(StopIteration, next, iterator)

    def test_split_by_empty(self):
        """Test split with empty iterable."""
        iterable = []

        def condition(item):
            return item % 3 == 0

        iterator = dnf.util.split_by(iterable, condition)

        self.assertEqual(next(iterator), ())
        self.assertRaises(StopIteration, next, iterator)

    def test_strip_prefix(self):
        self.assertIsNone(dnf.util.strip_prefix("razorblade", "blade"))
        self.assertEqual(dnf.util.strip_prefix("razorblade", "razor"), "blade")

    def test_touch(self):
        self.assertRaises(OSError, dnf.util.touch,
                          tests.support.NONEXISTENT_FILE, no_create=True)


class TestMultiCall(tests.support.TestCase):
    def test_multi_call(self):
        l = dnf.util.MultiCallList(["one", "two", "three"])
        self.assertEqual(l.upper(), ["ONE", "TWO", "THREE"])
        self.assertEqual(l.pop(), "three")

    def test_assignment(self):
        o1 = mock.Mock(x=3)
        o2 = mock.Mock(x=5)
        l = dnf.util.MultiCallList([o1, o2])
        l.x = 5
        self.assertEqual([5, 5], list(map(operator.attrgetter('x'), [o1, o2])))



# -*- coding: utf-8 -*-

# Copyright (C) 2012-2018 Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#

from __future__ import absolute_import
from __future__ import unicode_literals

import itertools

import dnf

import tests.support


class Reinstall(tests.support.ResultTestCase):

    REPOS = ['main', 'updates']

    def setUp(self):
        super(Reinstall, self).setUp()
        self.base.conf.multilib_policy = 'all'

    def test_package_reinstall(self):
        p = self.base.sack.query().available().filter(
            nevra="librita-1-1.x86_64")[0]
        self.assertEqual(1, self.base.package_reinstall(p))
        self.base.resolve()
        self.assertEqual(1, len(self.base._goal.list_reinstalls()))

    def test_package_reinstall_fail(self):
        p = self.base.sack.query().available().filter(nevra="hole-1-2.x86_64")[0]
        with self.assertRaises(dnf.exceptions.MarkingError) as context:
            self.base.package_reinstall(p)
        self.assertEqual(context.exception.pkg_spec, 'hole')
        self.base.resolve()
        self.assertEmpty(self.base._goal.list_downgrades())

    def test_reinstall(self):
        cnt = self.base.reinstall('pepper')
        self.assertEqual(cnt, 1)
        new_set = tests.support.installed_but(self.sack, "pepper")
        available_query = self.sack.query().available()
        new_set += list(available_query._nevra("pepper-20-0.x86_64"))
        self.assertResult(self.base, new_set)

    def test_reinstall_new_reponame_available(self):
        """Test whether it installs packages only from the repository."""
        reinstalled_count = self.base.reinstall('librita', new_reponame='main')

        self.assertEqual(reinstalled_count, 1)
        self.assertResult(self.base, itertools.chain(
            self.sack.query().installed().filter(name__neq='librita'),
            dnf.subject.Subject('librita.i686').get_best_query(self.sack).installed(),
            dnf.subject.Subject('librita').get_best_query(self.sack).available()))

    def test_reinstall_new_reponame_notavailable(self):
        """Test whether it installs packages only from the repository."""
        self.assertRaises(
            dnf.exceptions.PackagesNotAvailableError,
            self.base.reinstall, 'librita', new_reponame='non-main')

    def test_reinstall_new_reponame_neq_available(self):
        """Test whether it installs only packages not from the repository."""
        reinstalled_count = self.base.reinstall('librita', new_reponame_neq='non-main')

        self.assertEqual(reinstalled_count, 1)
        self.assertResult(self.base, itertools.chain(
            self.sack.query().installed().filter(name__neq='librita'),
            dnf.subject.Subject('librita.i686').get_best_query(self.sack).installed(),
            dnf.subject.Subject('librita').get_best_query(self.sack).available()))

    def test_reinstall_new_reponame_neq_notavailable(self):
        """Test whether it installs only packages not from the repository."""
        self.assertRaises(
            dnf.exceptions.PackagesNotAvailableError,
            self.base.reinstall, 'librita', new_reponame_neq='main')

    def test_reinstall_notfound(self):
        """Test whether it fails if the package does not exist."""
        with self.assertRaises(dnf.exceptions.PackagesNotInstalledError) as ctx:
            self.base.reinstall('non-existent')

        self.assertEqual(ctx.exception.pkg_spec, 'non-existent')
        self.assertResult(self.base, self.sack.query().installed())

    def test_reinstall_notinstalled(self):
        """Test whether it fails if the package is not installed."""
        with self.assertRaises(dnf.exceptions.PackagesNotInstalledError) as ctx:
            self.base.reinstall('lotus')

        self.assertEqual(ctx.exception.pkg_spec, 'lotus')
        self.assertResult(self.base, self.sack.query().installed())

    def test_reinstall_notavailable(self):
        """Test whether it fails if the package is not available."""
        with self.assertRaises(dnf.exceptions.PackagesNotAvailableError) as ctx:
            self.base.reinstall('hole')

        self.assertEqual(ctx.exception.pkg_spec, 'hole')
        self.assertCountEqual(
            ctx.exception.packages,
            dnf.subject.Subject('hole').get_best_query(self.sack).installed())
        self.assertResult(self.base, self.sack.query().installed())

    def test_reinstall_notavailable_available(self):
        """Test whether it does not fail if some packages are available and some not."""
        reinstalled_count = self.base.reinstall('librita')

        self.assertEqual(reinstalled_count, 1)
        self.assertResult(self.base, itertools.chain(
            self.sack.query().installed().filter(name__neq='librita'),
            dnf.subject.Subject('librita.i686').get_best_query(self.sack).installed(),
            dnf.subject.Subject('librita').get_best_query(self.sack).available()))

    def test_reinstall_old_reponame_installed(self):
        """Test whether it reinstalls packages only from the repository."""

        self.base = tests.support.MockBase('main')
        self.base.conf.multilib_policy = 'all'

        for pkg in self.base.sack.query().installed().filter(name='librita'):
            pkg._force_swdb_repoid = "main"
            self.history.rpm.add_install(pkg)
        self._swdb_commit()

        reinstalled_count = self.base.reinstall('librita', old_reponame='main')

        self.assertEqual(reinstalled_count, 1)
        self.assertResult(self.base, itertools.chain(
            self.sack.query().installed().filter(name__neq='librita'),
            dnf.subject.Subject('librita.i686').get_best_query(self.sack).installed(),
            dnf.subject.Subject('librita').get_best_query(self.sack).available())
        )

    def test_reinstall_old_reponame_notinstalled(self):
        """Test whether it reinstalls packages only from the repository."""
        self.assertRaises(
            dnf.exceptions.PackagesNotInstalledError,
            self.base.reinstall, 'librita', old_reponame='non-main'
        )

    def test_reinstall_remove_notavailable(self):
        """Test whether it removes the package which is not available."""
        self.base.reinstall('hole', remove_na=True)

        self.assertResult(
            self.base,
            self.sack.query().installed().filter(name__neq='hole')
        )



# -*- coding: utf-8 -*-

# Copyright (C) 2014-2018 Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#

from __future__ import absolute_import
from __future__ import unicode_literals

import logging

import dnf.cli.commands.remove
from dnf.cli.option_parser import OptionParser

import tests.support
from tests.support import mock


class RemoveCommandTest(tests.support.ResultTestCase):
    """Tests of ``dnf.cli.commands.EraseCommand`` class."""

    REPOS = []
    BASE_CLI = True
    CLI = "mock"

    def setUp(self):
        super(RemoveCommandTest, self).setUp()
        self.cmd = dnf.cli.commands.remove.RemoveCommand(self.base.mock_cli())

    def test_configure(self):
        parser = OptionParser()
        parser.parse_main_args(['autoremove', '-y'])
        parser.parse_command_args(self.cmd, ['autoremove', '-y'])
        self.cmd.configure()
        self.assertTrue(self.cmd.cli.demands.allow_erasing)

    @mock.patch('dnf.cli.commands.remove._',
                dnf.pycomp.NullTranslations().ugettext)
    def test_run_notfound(self):
        """Test whether it fails if the package cannot be found."""
        stdout = dnf.pycomp.StringIO()

        with tests.support.wiretap_logs('dnf', logging.INFO, stdout):
            tests.support.command_run(self.cmd, ['non-existent'])
        self.assertEqual(stdout.getvalue(),
                         'No match for argument: non-existent\nNo packages marked for removal.\n')
        self.assertResult(self.cmd.base, self.cmd.base.sack.query().installed())



# Copyright (C) 2015  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.

"""A plugin that ensures that given features are present."""


import dnf.cli
from dnf.i18n import _
from dnf.cli.option_parser import OptionParser

# The parent class allows registration to the CLI manager.
class Command(dnf.cli.Command):

    """A command that ensures that given features are present."""

    # An alias is needed to invoke the command from command line.
    aliases = ['foo']  # <-- SET YOUR ALIAS HERE.

    def configure(self):
        """Setup the demands."""
        # Repositories are needed if we want to install anything.
        self.cli.demands.available_repos = True
        # A sack is required by marking methods and dependency resolving.
        self.cli.demands.sack_activation = True
        # Resolving performs a transaction that installs the packages.
        self.cli.demands.resolving = True
        # Based on the system, privileges are required to do an installation.
        self.cli.demands.root_user = True  # <-- SET YOUR FLAG HERE.

    @staticmethod
    def set_argparser(parser):
        """Parse command line arguments."""
        parser.add_argument('package', nargs='+', metavar=_('PACKAGE'),
                            action=OptionParser.ParseSpecGroupFileCallback,
                            help=_('Package to install'))

    def run(self):
        """Run the command."""
        # Feature marking methods set the user request.
        for ftr_spec in self.opts.pkg_specs:
            try:
                self.base.install(ftr_spec)
            except dnf.exceptions.MarkingError:
                raise dnf.exceptions.Error('feature(s) not found: ' + ftr_spec)
        # Package marking methods set the user request.
        for pkg in self.base.add_remote_rpms(self.opts.filenames, strict=False):
            try:
                self.base.package_install(pkg, strict=False)
            except dnf.exceptions.MarkingError as e:
                raise dnf.exceptions.Error(e)
        # Comps data reading initializes the base.comps attribute.
        if self.opts.grp_specs:
            self.base.read_comps(arch_filter=True)
        # Group marking methods set the user request.
        for grp_spec in self.opts.grp_specs:
            group = self.base.comps.group_by_pattern(grp_spec)
            if not group:
                raise dnf.exceptions.Error('group not found: ' + grp_spec)
            self.base.group_install(group.id, ['mandatory', 'default'])


# Every plugin must be a subclass of dnf.Plugin.
class Plugin(dnf.Plugin):

    """A plugin that registers our custom command."""

    # Every plugin must provide its name.
    name = 'foo'  # <-- SET YOUR NAME HERE.

    # Every plugin must provide its own initialization function.
    def __init__(self, base, cli):
        """Initialize the plugin."""
        super(Plugin, self).__init__(base, cli)
        if cli:
            cli.register_command(Command)



# dnf configuration classes.
#
# Copyright (C) 2016-2017 Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#

from __future__ import absolute_import
from __future__ import unicode_literals

from dnf.yum import misc
from dnf.i18n import ucd, _
from dnf.pycomp import basestring, urlparse

import fnmatch
import dnf.conf.substitutions
import dnf.const
import dnf.exceptions
import dnf.pycomp
import dnf.util
import hawkey
import logging
import os
import libdnf.conf
import libdnf.repo
import tempfile

PRIO_EMPTY = libdnf.conf.Option.Priority_EMPTY
PRIO_DEFAULT = libdnf.conf.Option.Priority_DEFAULT
PRIO_MAINCONFIG = libdnf.conf.Option.Priority_MAINCONFIG
PRIO_AUTOMATICCONFIG = libdnf.conf.Option.Priority_AUTOMATICCONFIG
PRIO_REPOCONFIG = libdnf.conf.Option.Priority_REPOCONFIG
PRIO_PLUGINDEFAULT = libdnf.conf.Option.Priority_PLUGINDEFAULT
PRIO_PLUGINCONFIG = libdnf.conf.Option.Priority_PLUGINCONFIG
PRIO_COMMANDLINE = libdnf.conf.Option.Priority_COMMANDLINE
PRIO_RUNTIME = libdnf.conf.Option.Priority_RUNTIME

logger = logging.getLogger('dnf')


class BaseConfig(object):
    """Base class for storing configuration definitions.

       Subclass when creating your own definitions.

    """

    def __init__(self, config=None, section=None, parser=None):
        self.__dict__["_config"] = config
        self._section = section

    def __getattr__(self, name):
        if "_config" not in self.__dict__:
            raise AttributeError("'{}' object has no attribute '{}'".format(self.__class__, name))
        option = getattr(self._config, name)
        if option is None:
            return None
        try:
            value = option().getValue()
        except Exception as ex:
            return None
        if isinstance(value, str):
            return ucd(value)
        return value

    def __setattr__(self, name, value):
        option = getattr(self._config, name, None)
        if option is None:
            # unknown config option, store to BaseConfig only
            return super(BaseConfig, self).__setattr__(name, value)
        self._set_value(name, value, PRIO_RUNTIME)

    def __str__(self):
        out = []
        out.append('[%s]' % self._section)
        if self._config:
            for optBind in self._config.optBinds():
                try:
                    value = optBind.second.getValueString()
                except RuntimeError:
                    value = ""
                out.append('%s: %s' % (optBind.first, value))
        return '\n'.join(out)

    def _has_option(self, name):
        method = getattr(self._config, name, None)
        return method is not None

    def _get_value(self, name):
        method = getattr(self._config, name, None)
        if method is None:
            return None
        return method().getValue()

    def _get_priority(self, name):
        method = getattr(self._config, name, None)
        if method is None:
            return None
        return method().getPriority()

    def _set_value(self, name, value, priority=PRIO_RUNTIME):
        """Set option's value if priority is equal or higher
           than current priority."""
        method = getattr(self._config, name, None)
        if method is None:
            raise Exception("Option \"" + name + "\" does not exists")
        option = method()
        if value is None:
            try:
                option.set(priority, value)
            except Exception:
                pass
        else:
            try:
                if isinstance(value, list) or isinstance(value, tuple):
                    option.set(priority, libdnf.conf.VectorString(value))
                elif (isinstance(option, libdnf.conf.OptionBool)
                      or isinstance(option, libdnf.conf.OptionChildBool)
                      ) and isinstance(value, int):
                    option.set(priority, bool(value))
                else:
                    option.set(priority, value)
            except RuntimeError as e:
                raise dnf.exceptions.ConfigError(_("Error parsing '%s': %s")
                                                 % (value, str(e)),
                                                 raw_error=str(e))

    def _populate(self, parser, section, filename, priority=PRIO_DEFAULT):
        """Set option values from an INI file section."""
        if parser.hasSection(section):
            for name in parser.options(section):
                value = parser.getSubstitutedValue(section, name)
                if not value or value == 'None':
                    value = ''
                if hasattr(self._config, name):
                    try:
                        self._config.optBinds().at(name).newString(priority, value)
                    except RuntimeError as e:
                        logger.debug(_('Unknown configuration value: %s=%s in %s; %s'),
                                     ucd(name), ucd(value), ucd(filename), str(e))
                else:
                    if name == 'arch' and hasattr(self, name):
                        setattr(self, name, value)
                    else:
                        logger.debug(
                            _('Unknown configuration option: %s = %s in %s'),
                            ucd(name), ucd(value), ucd(filename))

    def dump(self):
        # :api
        """Return a string representing the values of all the
           configuration options.
        """
        output = ['[%s]' % self._section]

        if self._config:
            for optBind in self._config.optBinds():
                # if not opt._is_runtimeonly():
                try:
                    output.append('%s = %s' % (optBind.first, optBind.second.getValueString()))
                except RuntimeError:
                    pass

        return '\n'.join(output) + '\n'

    @staticmethod
    def write_raw_configfile(filename, section_id, substitutions, modify):
        # :api
        """
        filename   - name of config file (.conf or .repo)
        section_id - id of modified section (e.g. main, fedora, updates)
        substitutions - instance of base.conf.substitutions
        modify     - dict of modified options
        """
        parser = libdnf.conf.ConfigParser()
        parser.read(filename)

        # b/c repoids can have $values in them we need to map both ways to figure
        # out which one is which
        if not parser.hasSection(section_id):
            for sect in parser.getData():
                if libdnf.conf.ConfigParser.substitute(sect, substitutions) == section_id:
                    section_id = sect

        for name, value in modify.items():
            if isinstance(value, list):
                value = ' '.join(value)
            parser.setValue(section_id, name, value)

        parser.write(filename, False)


class MainConf(BaseConfig):
    # :api
    """Configuration option definitions for dnf.conf's [main] section."""
    def __init__(self, section='main', parser=None):
        # pylint: disable=R0915
        config = libdnf.conf.ConfigMain()
        super(MainConf, self).__init__(config, section, parser)
        self._set_value('pluginpath', [dnf.const.PLUGINPATH], PRIO_DEFAULT)
        self._set_value('pluginconfpath', [dnf.const.PLUGINCONFPATH], PRIO_DEFAULT)
        self.substitutions = dnf.conf.substitutions.Substitutions()
        self.arch = hawkey.detect_arch()
        self._config.system_cachedir().set(PRIO_DEFAULT, dnf.const.SYSTEM_CACHEDIR)

        # setup different cache and log for non-privileged users
        if dnf.util.am_i_root():
            cachedir = dnf.const.SYSTEM_CACHEDIR
            logdir = '/var/log'
        else:
            try:
                cachedir = logdir = misc.getCacheDir()
            except (IOError, OSError) as e:
                msg = _('Could not set cachedir: {}').format(ucd(e))
                raise dnf.exceptions.Error(msg)

        self._config.cachedir().set(PRIO_DEFAULT, cachedir)
        self._config.logdir().set(PRIO_DEFAULT, logdir)

        # track list of temporary files created
        self.tempfiles = []

    def __del__(self):
        for file_name in self.tempfiles:
            os.unlink(file_name)

    @property
    def get_reposdir(self):
        # :api
        """Returns the value of reposdir"""
        myrepodir = None
        # put repo file into first reposdir which exists or create it
        for rdir in self._get_value('reposdir'):
            if os.path.exists(rdir):
                myrepodir = rdir
                break

        if not myrepodir:
            myrepodir = self._get_value('reposdir')[0]
            dnf.util.ensure_dir(myrepodir)
        return myrepodir

    def _check_remote_file(self, optname):
        """
        In case the option value is a remote URL, download it to the temporary location
        and use this temporary file instead.
        """
        prio = self._get_priority(optname)
        val = self._get_value(optname)
        if isinstance(val, basestring):
            location = urlparse.urlparse(val)
            if location[0] in ('file', ''):
                # just strip the file:// prefix
                self._set_value(optname, location.path, prio)
            else:
                downloader = libdnf.repo.Downloader()
                temp_fd, temp_path = tempfile.mkstemp(prefix='dnf-downloaded-config-')
                self.tempfiles.append(temp_path)
                try:
                    downloader.downloadURL(None, val, temp_fd)
                except RuntimeError as e:
                    raise dnf.exceptions.ConfigError(
                        _('Configuration file URL "{}" could not be downloaded:\n'
                          '  {}').format(val, str(e)))
                else:
                    self._set_value(optname, temp_path, prio)
                finally:
                    os.close(temp_fd)

    def _search_inside_installroot(self, optname):
        """
        Return root used as prefix for option (installroot or "/"). When specified from commandline
        it returns value from conf.installroot
        """
        installroot = self._get_value('installroot')
        if installroot == "/":
            return installroot
        prio = self._get_priority(optname)
        # don't modify paths specified on commandline
        if prio >= PRIO_COMMANDLINE:
            return installroot
        val = self._get_value(optname)
        # if it exists inside installroot use it (i.e. adjust configuration)
        # for lists any component counts
        if not isinstance(val, str):
            if any(os.path.exists(os.path.join(installroot, p.lstrip('/'))) for p in val):
                self._set_value(
                    optname,
                    libdnf.conf.VectorString([self._prepend_installroot_path(p) for p in val]),
                    prio
                )
                return installroot
        elif os.path.exists(os.path.join(installroot, val.lstrip('/'))):
            self._set_value(optname, self._prepend_installroot_path(val), prio)
            return installroot
        return "/"

    def prepend_installroot(self, optname):
        # :api
        prio = self._get_priority(optname)
        new_path = self._prepend_installroot_path(self._get_value(optname))
        self._set_value(optname, new_path, prio)

    def _prepend_installroot_path(self, path):
        root_path = os.path.join(self._get_value('installroot'), path.lstrip('/'))
        return libdnf.conf.ConfigParser.substitute(root_path, self.substitutions)

    def _configure_from_options(self, opts):
        """Configure parts of CLI from the opts """
        config_args = ['plugins', 'version', 'config_file_path',
                       'debuglevel', 'errorlevel', 'installroot',
                       'best', 'assumeyes', 'assumeno', 'clean_requirements_on_remove', 'gpgcheck',
                       'showdupesfromrepos', 'plugins', 'ip_resolve',
                       'rpmverbosity', 'disable_excludes', 'color',
                       'downloadonly', 'exclude', 'excludepkgs', 'skip_broken',
                       'tsflags', 'arch', 'basearch', 'ignorearch', 'cacheonly', 'comment']

        for name in config_args:
            value = getattr(opts, name, None)
            if value is not None and value != []:
                if self._has_option(name):
                    appendValue = False
                    if self._config:
                        try:
                            appendValue = self._config.optBinds().at(name).getAddValue()
                        except RuntimeError:
                            # fails if option with "name" does not exist in _config (libdnf)
                            pass
                    if appendValue:
                        add_priority = dnf.conf.PRIO_COMMANDLINE
                        if add_priority < self._get_priority(name):
                            add_priority = self._get_priority(name)
                        for item in value:
                            if item:
                                self._set_value(name, self._get_value(name) + [item], add_priority)
                            else:
                                self._set_value(name, [], dnf.conf.PRIO_COMMANDLINE)
                    else:
                        self._set_value(name, value, dnf.conf.PRIO_COMMANDLINE)
                elif hasattr(self, name):
                    setattr(self, name, value)
                else:
                    logger.warning(_('Unknown configuration option: %s = %s'),
                                   ucd(name), ucd(value))

        if getattr(opts, 'gpgcheck', None) is False:
            self._set_value("localpkg_gpgcheck", False, dnf.conf.PRIO_COMMANDLINE)

        if hasattr(opts, 'main_setopts'):
            # now set all the non-first-start opts from main from our setopts
            # pylint: disable=W0212
            for name, values in opts.main_setopts.items():
                for val in values:
                    if hasattr(self._config, name):
                        try:
                            # values in main_setopts are strings, try to parse it using newString()
                            self._config.optBinds().at(name).newString(PRIO_COMMANDLINE, val)
                        except RuntimeError as e:
                            raise dnf.exceptions.ConfigError(
                                _("Error parsing --setopt with key '%s', value '%s': %s")
                                % (name, val, str(e)), raw_error=str(e))
                    else:
                        # if config option with "name" doesn't exist in _config, it could be defined
                        # only in Python layer
                        if hasattr(self, name):
                            setattr(self, name, val)
                        else:
                            msg = _("Main config did not have a %s attr. before setopt")
                            logger.warning(msg, name)

    def exclude_pkgs(self, pkgs):
        # :api
        name = "excludepkgs"

        if pkgs is not None and pkgs != []:
            if self._has_option(name):
                self._set_value(name, pkgs, dnf.conf.PRIO_COMMANDLINE)
            else:
                logger.warning(_('Unknown configuration option: %s = %s'),
                               ucd(name), ucd(pkgs))

    def _adjust_conf_options(self):
        """Adjust conf options interactions"""

        skip_broken_val = self._get_value('skip_broken')
        if skip_broken_val:
            self._set_value('strict', not skip_broken_val, self._get_priority('skip_broken'))

    @property
    def releasever(self):
        # :api
        return self.substitutions.get('releasever')

    @releasever.setter
    def releasever(self, val):
        # :api
        if val is None:
            self.substitutions.pop('releasever', None)
            return
        self.substitutions['releasever'] = str(val)

    @property
    def arch(self):
        # :api
        return self.substitutions.get('arch')

    @arch.setter
    def arch(self, val):
        # :api

        if val is None:
            self.substitutions.pop('arch', None)
            return
        if val not in dnf.rpm._BASEARCH_MAP.keys():
            msg = _('Incorrect or unknown "{}": {}')
            raise dnf.exceptions.Error(msg.format("arch", val))
        self.substitutions['arch'] = val
        self.basearch = dnf.rpm.basearch(val)

    @property
    def basearch(self):
        # :api
        return self.substitutions.get('basearch')

    @basearch.setter
    def basearch(self, val):
        # :api

        if val is None:
            self.substitutions.pop('basearch', None)
            return
        if val not in dnf.rpm._BASEARCH_MAP.values():
            msg = _('Incorrect or unknown "{}": {}')
            raise dnf.exceptions.Error(msg.format("basearch", val))
        self.substitutions['basearch'] = val

    def read(self, filename=None, priority=PRIO_DEFAULT):
        # :api
        if filename is None:
            filename = self._get_value('config_file_path')
        parser = libdnf.conf.ConfigParser()
        try:
            parser.read(filename)
        except RuntimeError as e:
            raise dnf.exceptions.ConfigError(_('Parsing file "%s" failed: %s') % (filename, e))
        except IOError as e:
            logger.warning(e)
        self._populate(parser, self._section, filename, priority)

        # update to where we read the file from
        self._set_value('config_file_path', filename, priority)

    @property
    def verbose(self):
        return self._get_value('debuglevel') >= dnf.const.VERBOSE_LEVEL


class RepoConf(BaseConfig):
    """Option definitions for repository INI file sections."""

    def __init__(self, parent, section=None, parser=None):
        masterConfig = parent._config if parent else libdnf.conf.ConfigMain()
        super(RepoConf, self).__init__(libdnf.conf.ConfigRepo(masterConfig), section, parser)
        # Do not remove! Attribute is a reference holder.
        # Prevents premature removal of the masterConfig. The libdnf ConfigRepo points to it.
        self._masterConfigRefHolder = masterConfig
        if section:
            self._config.name().set(PRIO_DEFAULT, section)

    def _configure_from_options(self, opts):
        """Configure repos from the opts. """

        if getattr(opts, 'gpgcheck', None) is False:
            for optname in ['gpgcheck', 'repo_gpgcheck']:
                self._set_value(optname, False, dnf.conf.PRIO_COMMANDLINE)

        repo_setopts = getattr(opts, 'repo_setopts', {})
        for repoid, setopts in repo_setopts.items():
            if not fnmatch.fnmatch(self._section, repoid):
                continue
            for name, values in setopts.items():
                for val in values:
                    if hasattr(self._config, name):
                        try:
                            # values in repo_setopts are strings, try to parse it using newString()
                            self._config.optBinds().at(name).newString(PRIO_COMMANDLINE, val)
                        except RuntimeError as e:
                            raise dnf.exceptions.ConfigError(
                                _("Error parsing --setopt with key '%s.%s', value '%s': %s")
                                % (self._section, name, val, str(e)), raw_error=str(e))
                    else:
                        msg = _("Repo %s did not have a %s attr. before setopt")
                        logger.warning(msg, self._section, name)



# -*- coding: utf-8 -*-

# Copyright (C) 2014-2018 Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#

from __future__ import absolute_import
from __future__ import unicode_literals

import os
from io import StringIO

import dnf.cli.cli

import tests.support
from tests.support import mock

'''
def _run(cli, args):
    with mock.patch('sys.stdout', new_callable=StringIO), \
            mock.patch('dnf.rpm.detect_releasever', return_value=69):
        cli.configure(['clean', '--config', '/dev/null'] + args)
        cli.run()


class CleanTest(tests.support.TestCase):
    def setUp(self):
        conf = dnf.conf.Conf()
        base = tests.support.Base(conf)
        base.repos.add(tests.support.MockRepo('main', conf))
        base.conf.reposdir = '/dev/null'
        base.conf.plugins = False
        base.output = tests.support.MockOutput()

        repo = base.repos['main']
        repo.baseurl = ['http:///dnf-test']
        repo.basecachedir = base.conf.cachedir

        walk = [
            (
                repo.basecachedir,
                [os.path.basename(repo._cachedir)],
                [repo.id + '.solv'],
            ),
            (repo._cachedir, ['repodata', 'packages'], ['metalink.xml']),
            (repo._cachedir + '/repodata', [], ['foo.xml', 'bar.xml.bz2']),
            (repo._cachedir + '/packages', [], ['foo.rpm']),
        ]
        os.walk = self.walk = mock.Mock(return_value=walk)
        self.base = base
        self.cli = dnf.cli.cli.Cli(base)

    def tearDown(self):
        self.base.close()

    def test_run(self):
        with mock.patch('dnf.cli.commands.clean._clean') as _clean:
            for args in [['all'],
                         ['metadata'],
                         ['metadata', 'packages'],
                         ['metadata', 'packages', 'expire-cache'],
                         ['dbcache'],
                         ['expire-cache']]:
                _run(self.cli, args)

        calls = [call[0] for call in _clean.call_args_list]
        counts = (5, 4, 5, 5, 1, 0)
        for call, count in zip(calls, counts):
            files = list(call[1])
            assert len(files) == count

    def test_walk_once(self):
        _run(self.cli, ['all'])
        assert len(self.walk.call_args_list) == 1

    def test_clean_local_repo(self):
        cachedir = self.base.conf.cachedir
        repo = self.base.repos['main']
        repo.baseurl = ['file:///localrepo']

        _run(self.cli, ['all'])

        # Make sure we never looked outside the base cachedir
        dirs = [call[0][0] for call in self.walk.call_args_list]
        assert all(d == cachedir for d in dirs)
'''



# -*- coding: utf-8 -*-

# Copyright (C) 2012-2018 Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#

from __future__ import absolute_import
from __future__ import unicode_literals

import dnf.cli.aliases
import os

import tests.support

CONF = tests.support.resource_path('etc/aliases.d/aliases.conf')


class AliasesTest(tests.support.TestCase):

    def setUp(self):
        # Set DNF_ALIASES_DISABLED so that no default config is loaded
        self.env_var_set = False
        if 'DNF_ALIASES_DISABLED' not in os.environ:
            os.environ['DNF_ALIASES_DISABLED'] = '1'
            self.env_var_set = True

        self.aliases_base = dnf.cli.aliases.Aliases()
        self.aliases_base._load_aliases([CONF])

    def tearDown(self):
        if self.env_var_set:
            del os.environ['DNF_ALIASES_DISABLED']

    def test_undefined(self):
        args = ['undefined', 'undefined2']
        expected_args = list(args)
        args = self.aliases_base._resolve(args)
        self.assertEqual(args, expected_args)

    def test_simple(self):
        args = ['h']
        expected_args = ['history']
        args = self.aliases_base._resolve(args)
        self.assertEqual(args, expected_args)

    def test_command_changed(self):
        args = ['cu']
        expected_args = ['check-update']
        args = self.aliases_base._resolve(args)
        self.assertEqual(args, expected_args)

    def test_package_unchanged(self):
        args = ['install', 'cu']
        expected_args = ['install', 'cu']
        args = self.aliases_base._resolve(args)
        self.assertEqual(args, expected_args)

    def test_recursive(self):
        args = ['lsi']
        expected_args = ['list', 'installed']
        args = self.aliases_base._resolve(args)
        self.assertEqual(args, expected_args)

    def test_options(self):
        args = ['FORCE']
        expected_args = ['--skip-broken', '--disableexcludes=all']
        args = self.aliases_base._resolve(args)
        self.assertEqual(args, expected_args)

    def test_options_recursive(self):
        args = ['force-inst']
        expected_args = ['--skip-broken', '--disableexcludes=all', 'install']
        args = self.aliases_base._resolve(args)
        self.assertEqual(args, expected_args)



# -*- coding: utf-8 -*-

# Copyright (C) 2016-2018 Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import dnf.cli.commands.check
import dnf.pycomp

import tests.support


EXPECTED_DUPLICATES_FORMAT = """\
dup-1-0.noarch is a duplicate with dup-2-0.noarch
dup-1-0.noarch is a duplicate with dup-3-0.noarch
"""

EXPECTED_OBSOLETED_FORMAT = """\
test-1-0.noarch is obsoleted by obs-3-0.noarch
"""


class CheckDuplicatesTest(tests.support.DnfBaseTestCase):

    REPOS = []
    BASE_CLI = True
    CLI = "stub"

    def test_duplicates(self):
        self.cmd = dnf.cli.commands.check.CheckCommand(self.cli)
        tests.support.command_configure(self.cmd, ['--duplicates'])
        with tests.support.patch_std_streams() as (stdout, _):
            with self.assertRaises(dnf.exceptions.Error) as ctx:
                self.cmd.run()
            self.assertEqual(str(ctx.exception),
                             'Check discovered 2 problem(s)')
        self.assertEqual(stdout.getvalue(), EXPECTED_DUPLICATES_FORMAT)

    def test_obsoleted(self):
        self.cmd = dnf.cli.commands.check.CheckCommand(self.cli)
        tests.support.command_configure(self.cmd, ['--obsoleted'])
        with tests.support.patch_std_streams() as (stdout, _):
            with self.assertRaises(dnf.exceptions.Error) as ctx:
                self.cmd.run()
            self.assertEqual(str(ctx.exception),
                             'Check discovered 1 problem(s)')
        self.assertEqual(stdout.getvalue(), EXPECTED_OBSOLETED_FORMAT)
